from __future__ import annotations

import copy
import datetime
import inspect
import itertools
import json
import logging
import traceback
import warnings
from collections import defaultdict
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import pandas as pd
from marshmallow import ValidationError

from great_expectations import __version__ as ge_version
from great_expectations._docs_decorators import deprecated_argument
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.expectation_suite import (
    ExpectationSuite,
    expectationSuiteSchema,
)
from great_expectations.core.expectation_validation_result import (
    ExpectationSuiteValidationResult,
    ExpectationValidationResult,
)
from great_expectations.core.run_identifier import RunIdentifier
from great_expectations.data_context.data_context.context_factory import project_manager
from great_expectations.data_context.types.base import CheckpointValidationDefinition
from great_expectations.exceptions import (
    GreatExpectationsError,
    InvalidExpectationConfigurationError,
)
from great_expectations.execution_engine.pandas_batch_data import PandasBatchData
from great_expectations.expectations.expectation_configuration import (
    ExpectationConfiguration,
)
from great_expectations.expectations.registry import (
    get_expectation_impl,
    list_registered_expectation_implementations,
)
from great_expectations.util import convert_to_json_serializable  # noqa: TID251 # FIXME CoP
from great_expectations.validator.exception_info import ExceptionInfo
from great_expectations.validator.metrics_calculator import (
    MetricsCalculator,
    _AbortedMetricsInfoDict,
    _MetricsDict,
)
from great_expectations.validator.util import recursively_convert_to_json_serializable
from great_expectations.validator.validation_graph import (
    ExpectationValidationGraph,
    MetricEdge,
    ValidationGraph,
)
from great_expectations.validator.validation_statistics import (
    calc_validation_statistics,
)

logger = logging.getLogger(__name__)
logging.captureWarnings(True)


if TYPE_CHECKING:
    from great_expectations.core.batch import (
        AnyBatch,
        Batch,
        BatchDataUnion,
        BatchMarkers,
        LegacyBatchDefinition,
    )
    from great_expectations.core.id_dict import BatchSpec
    from great_expectations.data_context.data_context import AbstractDataContext
    from great_expectations.datasource.fluent.interfaces import Batch as FluentBatch
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.validator.metric_configuration import MetricConfiguration


@dataclass
class ValidationDependencies:
    # Note: Dependent "metric_name" (key) is different from "metric_name" in dependency "MetricConfiguration" (value).  # noqa: E501 # FIXME CoP
    metric_configurations: Dict[str, MetricConfiguration] = field(default_factory=dict)
    result_format: Dict[str, Any] = field(default_factory=dict)

    def set_metric_configuration(
        self, metric_name: str, metric_configuration: MetricConfiguration
    ) -> None:
        """
        Sets specified "MetricConfiguration" for "metric_name" to "metric_configurations" dependencies dictionary.
        """  # noqa: E501 # FIXME CoP
        self.metric_configurations[metric_name] = metric_configuration

    def get_metric_configuration(self, metric_name: str) -> Optional[MetricConfiguration]:
        """
        Obtains "MetricConfiguration" for specified "metric_name" from "metric_configurations" dependencies dictionary.
        """  # noqa: E501 # FIXME CoP
        return self.metric_configurations.get(metric_name)

    def remove_metric_configuration(self, metric_name: str) -> None:
        """
        Removes "MetricConfiguration" for specified "metric_name" from "metric_configurations" dependencies dictionary.
        """  # noqa: E501 # FIXME CoP
        del self.metric_configurations[metric_name]

    def get_metric_names(self) -> List[str]:
        """
        Returns "metric_name" keys, for which "MetricConfiguration" dependency objects have been specified.
        """  # noqa: E501 # FIXME CoP
        return list(self.metric_configurations.keys())

    def get_metric_configurations(self) -> List[MetricConfiguration]:
        """
        Returns "MetricConfiguration" dependency objects specified.
        """
        return list(self.metric_configurations.values())


class Validator:
    """Validator is the key object used to create Expectations, validate Expectations, and get Metrics for Expectations.

    Validators are used by Checkpoints to validate Expectations.

    Args:
        execution_engine: The Execution Engine to be used to perform validation.
        interactive_evaluation: If True, the Validator will perform evaluation when Expectations are added.
        expectation_suite: The Expectation Suite to validate.
        expectation_suite_name: The name of the Expectation Suite to validate.
        data_context: The Data Context associated with this Validator.
        batches: The Batches for which to validate.
    """  # noqa: E501 # FIXME CoP

    DEFAULT_RUNTIME_CONFIGURATION = {
        "catch_exceptions": False,
        "result_format": "BASIC",
    }
    RUNTIME_KEYS = DEFAULT_RUNTIME_CONFIGURATION.keys()

    # noinspection PyUnusedLocal
    def __init__(  # noqa: PLR0913 # FIXME CoP
        self,
        execution_engine: ExecutionEngine,
        interactive_evaluation: bool = True,
        expectation_suite: ExpectationSuite | None = None,
        expectation_suite_name: Optional[str] = None,
        data_context: Optional[AbstractDataContext] = None,
        batches: List[Batch] | Sequence[Batch | FluentBatch] = tuple(),
        **kwargs,
    ) -> None:
        self._data_context: Optional[AbstractDataContext] = data_context

        self._metrics_calculator: MetricsCalculator = MetricsCalculator(
            execution_engine=execution_engine,
            show_progress_bars=self._determine_progress_bars(),
        )
        execution_engine.batch_manager.reset_batch_cache()
        self._execution_engine: ExecutionEngine = execution_engine

        if batches:
            self.load_batch_list(batch_list=batches)

        self._expose_dataframe_methods: bool = False

        self.interactive_evaluation: bool = interactive_evaluation
        self._initialize_expectations(
            expectation_suite=expectation_suite,
            expectation_suite_name=expectation_suite_name,
        )
        self._default_expectation_args: Dict[str, Union[bool, str]] = copy.deepcopy(
            Validator.DEFAULT_RUNTIME_CONFIGURATION  # type: ignore[arg-type] # FIXME CoP
        )

        # This special state variable tracks whether a validation run is going on, which will disable  # noqa: E501 # FIXME CoP
        # saving expectation config objects
        self._active_validation: bool = False

    @property
    def _include_rendered_content(self) -> bool:
        return project_manager.is_using_cloud()

    @property
    def execution_engine(self) -> ExecutionEngine:
        """Returns the execution engine being used by the validator at the given time"""
        return self._execution_engine

    @property
    def metrics_calculator(self) -> MetricsCalculator:
        """Returns the "MetricsCalculator" object being used by the Validator to handle metrics computations."""  # noqa: E501 # FIXME CoP
        return self._metrics_calculator

    @property
    def data_context(self) -> Optional[AbstractDataContext]:
        """Reference to DataContext object handle."""
        return self._data_context

    @property
    def expose_dataframe_methods(self) -> bool:
        """The "expose_dataframe_methods" getter property."""
        return self._expose_dataframe_methods

    @expose_dataframe_methods.setter
    def expose_dataframe_methods(self, value: bool) -> None:
        """The "expose_dataframe_methods" setter property."""
        self._expose_dataframe_methods = value

    @property
    def loaded_batch_ids(self) -> List[str]:
        """Getter for IDs of loaded Batch objects (convenience property)"""
        return self._execution_engine.batch_manager.loaded_batch_ids

    @property
    def active_batch_data(self) -> Optional[BatchDataUnion]:
        """Getter for BatchData object from the currently-active Batch object (convenience property)."""  # noqa: E501 # FIXME CoP
        return self._execution_engine.batch_manager.active_batch_data

    @property
    def batch_cache(self) -> Dict[str, AnyBatch]:
        """Getter for dictionary of Batch objects (convenience property)"""
        return self._execution_engine.batch_manager.batch_cache

    @property
    def batches(self) -> Dict[str, AnyBatch]:
        """Getter for dictionary of Batch objects (alias convenience property, to be deprecated)"""
        return self.batch_cache

    @property
    def active_batch_id(self) -> Optional[str]:
        """Getter for batch_id of active Batch (convenience property)"""
        return self._execution_engine.batch_manager.active_batch_id

    @property
    def active_batch(self) -> Optional[AnyBatch]:
        """Getter for active Batch (convenience property)"""
        return self._execution_engine.batch_manager.active_batch

    @property
    def active_batch_spec(self) -> Optional[BatchSpec]:
        """Getter for batch_spec of active Batch (convenience property)"""
        return self._execution_engine.batch_manager.active_batch_spec

    @property
    def active_batch_markers(self) -> Optional[BatchMarkers]:
        """Getter for batch_markers of active Batch (convenience property)"""
        return self._execution_engine.batch_manager.active_batch_markers

    @property
    def active_batch_definition(self) -> Optional[LegacyBatchDefinition]:
        """Getter for batch_definition of active Batch (convenience property)"""
        return self._execution_engine.batch_manager.active_batch_definition

    @property
    def expectation_suite(self) -> ExpectationSuite:
        return self._expectation_suite

    @expectation_suite.setter
    def expectation_suite(self, value: ExpectationSuite) -> None:
        self._initialize_expectations(
            expectation_suite=value,
            expectation_suite_name=value.name,
        )

    @property
    def expectation_suite_name(self) -> str:
        """Gets the current expectation_suite name of this data_asset as stored in the expectations configuration."""  # noqa: E501 # FIXME CoP
        return self._expectation_suite.name

    @expectation_suite_name.setter
    def expectation_suite_name(self, name: str) -> None:
        """Sets the expectation_suite name of this data_asset as stored in the expectations configuration."""  # noqa: E501 # FIXME CoP
        self._expectation_suite.name = name

    def load_batch_list(self, batch_list: Sequence[Batch | FluentBatch]) -> None:
        self._execution_engine.batch_manager.load_batch_list(batch_list=batch_list)

    def get_metric(
        self,
        metric: MetricConfiguration,
    ) -> Any:
        """Convenience method, return the value of the requested metric.

        Args:
            metric: MetricConfiguration

        Returns:
            The value of the requested metric.
        """
        return self._metrics_calculator.get_metric(metric=metric)

    def get_metrics(
        self,
        metrics: Dict[str, MetricConfiguration],
    ) -> Dict[str, Any]:
        """
        Convenience method that resolves requested metrics (specified as dictionary, keyed by MetricConfiguration ID).

        Args:
            metrics: Dictionary of desired metrics to be resolved; metric_name is key and MetricConfiguration is value.

        Returns:
            Return Dictionary with requested metrics resolved, with metric_name as key and computed metric as value.
        """  # noqa: E501 # FIXME CoP
        return self._metrics_calculator.get_metrics(metrics=metrics)

    def compute_metrics(
        self,
        metric_configurations: List[MetricConfiguration],
        runtime_configuration: Optional[dict] = None,
        min_graph_edges_pbar_enable: int = 0,
        # Set to low number (e.g., 3) to suppress progress bar for small graphs.
    ) -> tuple[_MetricsDict, _AbortedMetricsInfoDict]:
        """
        Convenience method that computes requested metrics (specified as elements of "MetricConfiguration" list).

        Args:
            metric_configurations: List of desired MetricConfiguration objects to be resolved.
            runtime_configuration: Additional run-time settings (see "Validator.DEFAULT_RUNTIME_CONFIGURATION").
            min_graph_edges_pbar_enable: Minumum number of graph edges to warrant showing progress bars.

        Returns:
            Tuple of two elements, the first is a dictionary with requested metrics resolved,
            with unique metric ID as key and computed metric as value. The second is a dictionary of the
            aborted metrics information, with metric ID as key if any metrics were aborted.
        """  # noqa: E501 # FIXME CoP
        return self._metrics_calculator.compute_metrics(
            metric_configurations=metric_configurations,
            runtime_configuration=runtime_configuration,
            min_graph_edges_pbar_enable=min_graph_edges_pbar_enable,
        )

    def columns(self, domain_kwargs: Optional[Dict[str, Any]] = None) -> List[str]:
        """Convenience method to obtain Batch columns.

        Arguments:
            domain_kwargs: Optional dictionary of domain kwargs (e.g., containing "batch_id").

        Returns:
            The list of Batch columns.
        """
        return self._metrics_calculator.columns(domain_kwargs=domain_kwargs)

    def head(
        self,
        n_rows: int = 5,
        domain_kwargs: Optional[Dict[str, Any]] = None,
        fetch_all: bool = False,
    ) -> pd.DataFrame:
        """Convenience method to return the first several rows or records from a Batch of data.

        Args:
            n_rows: The number of rows to return.
            domain_kwargs: If provided, the domain for which to return records.
            fetch_all: If True, ignore n_rows and return the entire batch.

        Returns:
            A Pandas DataFrame containing the records' data.
        """
        return self._metrics_calculator.head(
            n_rows=n_rows, domain_kwargs=domain_kwargs, fetch_all=fetch_all
        )

    @override
    def __dir__(self) -> List[str]:
        """
        This custom magic method is used to enable expectation tab completion on Validator objects.
        It also allows users to call Pandas.DataFrame methods on Validator objects
        """
        validator_attrs = set(super().__dir__())
        class_expectation_impls = set(list_registered_expectation_implementations())
        # execution_engine_expectation_impls = (
        #     {
        #         attr_name
        #         for attr_name in self.execution_engine.__dir__()
        #         if attr_name.startswith("expect_")
        #     }
        #     if self.execution_engine
        #     else set()
        # )

        combined_dir = (
            validator_attrs | class_expectation_impls
            # | execution_engine_expectation_impls
        )

        if self._expose_dataframe_methods:
            combined_dir | set(dir(pd.DataFrame))

        return list(combined_dir)

    def _determine_progress_bars(self) -> bool:
        enable: bool = True
        if self._data_context:
            progress_bars = self._data_context.progress_bars
            # If progress_bars are not present, assume we want them enabled
            if progress_bars is not None:
                if "globally" in progress_bars:
                    enable = bool(progress_bars["globally"])
                if "metric_calculations" in progress_bars:
                    enable = bool(progress_bars["metric_calculations"])

        return enable

    def __getattr__(self, name):
        if self.active_batch is None:
            raise TypeError("active_batch cannot be None")  # noqa: TRY003 # FIXME CoP
        name = name.lower()
        if (
            name.startswith("expect_") or name == "unexpected_rows_expectation"
        ) and get_expectation_impl(name):
            return self.validate_expectation(name)
        elif (
            self._expose_dataframe_methods
            and isinstance(self.active_batch.data, PandasBatchData)
            and hasattr(pd.DataFrame, name)
        ):
            return getattr(self.active_batch.data.dataframe, name)
        else:
            raise AttributeError(f"'{type(self).__name__}'  object has no attribute '{name}'")  # noqa: TRY003 # FIXME CoP

    def validate_expectation(self, name: str) -> Callable:  # noqa: C901, PLR0915 # FIXME CoP
        """
        Given the name of an Expectation, obtains the Class-first Expectation implementation and utilizes the
                expectation's validate method to obtain a validation result. Also adds in the runtime configuration

                        Args:
                            name (str): The name of the Expectation being validated

                        Returns:
                            The Expectation's validation result
        """  # noqa: E501 # FIXME CoP
        expectation_impl = get_expectation_impl(name)

        def inst_expectation(*args: dict, **kwargs):  # noqa: C901, PLR0912 # FIXME CoP
            # this is used so that exceptions are caught appropriately when they occur in expectation config  # noqa: E501 # FIXME CoP

            # TODO: JPC - THIS LOGIC DOES NOT RESPECT DEFAULTS SET BY USERS IN THE VALIDATOR VS IN THE EXPECTATION  # noqa: E501 # FIXME CoP
            # DEVREL has action to develop a new plan in coordination with MarioPod

            expectation_kwargs = recursively_convert_to_json_serializable(kwargs)

            meta: Optional[dict] = expectation_kwargs.pop("meta", None)

            basic_default_expectation_args: dict = {
                k: v
                for k, v in self.default_expectation_args.items()
                if k in Validator.RUNTIME_KEYS
            }
            basic_runtime_configuration: dict = copy.deepcopy(basic_default_expectation_args)
            basic_runtime_configuration.update(
                {k: v for k, v in kwargs.items() if k in Validator.RUNTIME_KEYS}
            )

            allowed_config_keys: Tuple[str, ...] = expectation_impl.get_allowed_config_keys()

            args_keys: Tuple[str, ...] = expectation_impl.args_keys or tuple()

            arg_name: str

            idx: int
            arg: dict
            for idx, arg in enumerate(args):
                try:
                    arg_name = args_keys[idx]
                    if arg_name in allowed_config_keys:
                        expectation_kwargs[arg_name] = arg
                    if arg_name == "meta":
                        logger.warning(
                            "Setting meta via args could be ambiguous; please use a kwarg instead."
                        )
                        meta = arg
                except IndexError:
                    raise InvalidExpectationConfigurationError(  # noqa: TRY003 # FIXME CoP
                        f"Invalid positional argument: {arg}"
                    )

            configuration: ExpectationConfiguration | None = None

            try:
                expectation = expectation_impl(**expectation_kwargs, meta=meta)
                configuration = expectation.configuration

                if self.interactive_evaluation:
                    configuration.process_suite_parameters(
                        self._expectation_suite.suite_parameters,
                        True,
                        self._data_context,
                    )

                """Given an implementation and a configuration for any Expectation, returns its validation result"""  # noqa: E501 # FIXME CoP

                if not self.interactive_evaluation and not self._active_validation:
                    validation_result = ExpectationValidationResult(
                        expectation_config=copy.deepcopy(configuration)
                    )
                else:
                    validation_result = expectation.validate_(
                        validator=self,
                        suite_parameters=self._expectation_suite.suite_parameters,
                        data_context=self._data_context,
                        runtime_configuration=basic_runtime_configuration,
                    )

                # If validate has set active_validation to true, then we do not save the config to avoid  # noqa: E501 # FIXME CoP
                # saving updating expectation configs to the same suite during validation runs
                if self._active_validation is True:
                    stored_config = configuration.get_raw_configuration()
                else:
                    # Append the expectation to the config.
                    stored_config = self._expectation_suite._add_expectation(
                        expectation_configuration=configuration.get_raw_configuration(),
                    )

                # If there was no interactive evaluation, success will not have been computed.
                if validation_result.success is not None:
                    # Add a "success" object to the config
                    stored_config.success_on_last_run = validation_result.success

            except Exception as err:
                if basic_runtime_configuration.get("catch_exceptions"):
                    exception_traceback = traceback.format_exc()
                    exception_message = f"{type(err).__name__}: {err!s}"
                    exception_info = ExceptionInfo(
                        exception_traceback=exception_traceback,
                        exception_message=exception_message,
                    )

                    if not configuration:
                        configuration = ExpectationConfiguration(
                            type=name, kwargs=expectation_kwargs, meta=meta
                        )

                    validation_result = ExpectationValidationResult(
                        success=False,
                        exception_info=exception_info,
                        expectation_config=configuration,
                    )
                else:
                    raise err  # noqa: TRY201 # FIXME CoP

            if self._include_rendered_content:
                validation_result.render()

            return validation_result

        inst_expectation.__name__ = name
        inst_expectation.__doc__ = expectation_impl.__doc__

        return inst_expectation

    def list_available_expectation_types(self) -> List[str]:
        """Returns a list of all expectations available to the validator"""
        keys = dir(self)
        return [expectation for expectation in keys if expectation.startswith("expect_")]

    def graph_validate(
        self,
        configurations: List[ExpectationConfiguration],
        runtime_configuration: Optional[dict] = None,
    ) -> List[ExpectationValidationResult]:
        """Obtains validation dependencies for each metric using the implementation of their
        associated expectation, then proceeds to add these dependencies to the validation graph,
        supply readily available metric implementations to fulfill current metric requirements,
        and validate these metrics.

        Args:
            configurations(List[ExpectationConfiguration]): A list of needed Expectation
            Configurations that will be used to supply domain and values for metrics.
            runtime_configuration (dict): A dictionary of runtime keyword arguments, controlling
            semantics, such as the result_format.

        Returns:
            A list of Validations, validating that all necessary metrics are available.
        """
        if runtime_configuration is None:
            runtime_configuration = {}

        if runtime_configuration.get("catch_exceptions", True):
            catch_exceptions = True
        else:
            catch_exceptions = False

        expectation_validation_graphs: List[ExpectationValidationGraph]

        evrs: List[ExpectationValidationResult]

        processed_configurations: List[ExpectationConfiguration] = []

        (
            expectation_validation_graphs,
            evrs,
            processed_configurations,
        ) = self._generate_metric_dependency_subgraphs_for_each_expectation_configuration(
            expectation_configurations=configurations,
            processed_configurations=processed_configurations,
            catch_exceptions=catch_exceptions,
            runtime_configuration=runtime_configuration,
        )

        graph: ValidationGraph = self._generate_suite_level_graph_from_expectation_level_sub_graphs(
            expectation_validation_graphs=expectation_validation_graphs
        )

        resolved_metrics: _MetricsDict

        try:
            (
                resolved_metrics,
                evrs,
                processed_configurations,
            ) = self._resolve_suite_level_graph_and_process_metric_evaluation_errors(
                graph=graph,
                runtime_configuration=runtime_configuration,
                expectation_validation_graphs=expectation_validation_graphs,
                evrs=evrs,
                processed_configurations=processed_configurations,
                show_progress_bars=self._determine_progress_bars(),
            )
        except Exception as err:
            # If a general Exception occurs during the execution of "ValidationGraph.resolve()", then  # noqa: E501 # FIXME CoP
            # all expectations in the suite are impacted, because it is impossible to attribute the failure to a metric.  # noqa: E501 # FIXME CoP
            if catch_exceptions:
                exception_traceback: str = traceback.format_exc()
                evrs = self._catch_exceptions_in_failing_expectation_validations(
                    exception_traceback=exception_traceback,
                    exception=err,
                    failing_expectation_configurations=processed_configurations,
                    evrs=evrs,
                )
                return evrs
            else:
                raise err  # noqa: TRY201 # FIXME CoP

        configuration: ExpectationConfiguration
        result: ExpectationValidationResult
        for configuration in processed_configurations:
            try:
                runtime_configuration_default = copy.deepcopy(runtime_configuration)

                expectation = configuration.to_domain_obj()
                result = expectation.metrics_validate(
                    metrics=resolved_metrics,
                    execution_engine=self._execution_engine,
                    runtime_configuration=runtime_configuration_default,
                )
                evrs.append(result)
            except Exception as err:
                if catch_exceptions:
                    exception_traceback = traceback.format_exc()
                    evrs = self._catch_exceptions_in_failing_expectation_validations(
                        exception_traceback=exception_traceback,
                        exception=err,
                        failing_expectation_configurations=[configuration],
                        evrs=evrs,
                    )
                else:
                    raise err  # noqa: TRY201 # FIXME CoP

        return evrs

    def _generate_metric_dependency_subgraphs_for_each_expectation_configuration(
        self,
        expectation_configurations: List[ExpectationConfiguration],
        processed_configurations: List[ExpectationConfiguration],
        catch_exceptions: bool,
        runtime_configuration: Optional[dict] = None,
    ) -> Tuple[
        List[ExpectationValidationGraph],
        List[ExpectationValidationResult],
        List[ExpectationConfiguration],
    ]:
        # While evaluating expectation configurations, create sub-graph for every metric dependency and incorporate  # noqa: E501 # FIXME CoP
        # these sub-graphs under corresponding expectation-level sub-graph (state of ExpectationValidationGraph object).  # noqa: E501 # FIXME CoP
        expectation_validation_graphs: List[ExpectationValidationGraph] = []
        evrs: List[ExpectationValidationResult] = []
        configuration: ExpectationConfiguration
        evaluated_config: ExpectationConfiguration
        metric_configuration: MetricConfiguration
        graph: ValidationGraph
        for configuration in expectation_configurations:
            # Validating
            try:
                assert configuration.type is not None, (
                    "Given configuration should include expectation type"
                )
            except AssertionError as e:
                raise InvalidExpectationConfigurationError(str(e))

            evaluated_config = copy.deepcopy(configuration)

            if self.active_batch_id:
                evaluated_config.kwargs.update({"batch_id": self.active_batch_id})

            expectation = evaluated_config.to_domain_obj()
            validation_dependencies: ValidationDependencies = (
                expectation.get_validation_dependencies(
                    execution_engine=self._execution_engine,
                    runtime_configuration=runtime_configuration,
                )
            )

            try:
                expectation_validation_graph: ExpectationValidationGraph = ExpectationValidationGraph(  # noqa: E501 # FIXME CoP
                    configuration=evaluated_config,
                    graph=self._metrics_calculator.build_metric_dependency_graph(
                        metric_configurations=validation_dependencies.get_metric_configurations(),
                        runtime_configuration=runtime_configuration,
                    ),
                )
                expectation_validation_graphs.append(expectation_validation_graph)
                processed_configurations.append(evaluated_config)
            except Exception as err:
                if catch_exceptions:
                    exception_traceback: str = traceback.format_exc()
                    exception_message: str = str(err)
                    exception_info = ExceptionInfo(
                        exception_traceback=exception_traceback,
                        exception_message=exception_message,
                    )
                    result = ExpectationValidationResult(
                        success=False,
                        exception_info=exception_info,
                        expectation_config=evaluated_config,
                    )
                    evrs.append(result)
                else:
                    raise err  # noqa: TRY201 # FIXME CoP

        return expectation_validation_graphs, evrs, processed_configurations

    def _generate_suite_level_graph_from_expectation_level_sub_graphs(
        self,
        expectation_validation_graphs: List[ExpectationValidationGraph],
    ) -> ValidationGraph:
        # Collect edges from all expectation-level sub-graphs and incorporate them under common suite-level graph.  # noqa: E501 # FIXME CoP
        expectation_validation_graph: ExpectationValidationGraph
        edges: List[MetricEdge] = list(
            itertools.chain.from_iterable(
                [
                    expectation_validation_graph.graph.edges
                    for expectation_validation_graph in expectation_validation_graphs
                ]
            )
        )
        validation_graph = ValidationGraph(execution_engine=self._execution_engine, edges=edges)
        return validation_graph

    def _resolve_suite_level_graph_and_process_metric_evaluation_errors(  # noqa: PLR0913 # FIXME CoP
        self,
        graph: ValidationGraph,
        runtime_configuration: dict,
        expectation_validation_graphs: List[ExpectationValidationGraph],
        evrs: List[ExpectationValidationResult],
        processed_configurations: List[ExpectationConfiguration],
        show_progress_bars: bool,
    ) -> Tuple[
        _MetricsDict,
        List[ExpectationValidationResult],
        List[ExpectationConfiguration],
    ]:
        # Resolve overall suite-level graph and process any MetricResolutionError type exceptions that might occur.  # noqa: E501 # FIXME CoP
        resolved_metrics: _MetricsDict
        aborted_metrics_info: _AbortedMetricsInfoDict
        (
            resolved_metrics,
            aborted_metrics_info,
        ) = self._metrics_calculator.resolve_validation_graph(
            graph=graph,
            runtime_configuration=runtime_configuration,
            min_graph_edges_pbar_enable=0,
        )

        # Trace MetricResolutionError occurrences to expectations relying on corresponding malfunctioning metrics.  # noqa: E501 # FIXME CoP
        rejected_configurations: List[ExpectationConfiguration] = []
        for expectation_validation_graph in expectation_validation_graphs:
            metric_exception_info: Dict[str, Union[MetricConfiguration, ExceptionInfo, int]] = (
                expectation_validation_graph.get_exception_info(metric_info=aborted_metrics_info)
            )
            # Report all MetricResolutionError occurrences impacting expectation and append it to rejected list.  # noqa: E501 # FIXME CoP
            if len(metric_exception_info) > 0:
                configuration = expectation_validation_graph.configuration
                result = ExpectationValidationResult(
                    success=False,
                    exception_info=metric_exception_info,
                    expectation_config=configuration,
                )
                evrs.append(result)

                if configuration not in rejected_configurations:
                    rejected_configurations.append(configuration)

        # Exclude all rejected expectations from list of expectations cleared for validation.
        for configuration in rejected_configurations:
            processed_configurations.remove(configuration)

        return resolved_metrics, evrs, processed_configurations

    @staticmethod
    def _catch_exceptions_in_failing_expectation_validations(
        exception_traceback: str,
        exception: Exception,
        failing_expectation_configurations: List[ExpectationConfiguration],
        evrs: List[ExpectationValidationResult],
    ) -> List[ExpectationValidationResult]:
        """
        Catch exceptions in failing Expectation validations and convert to unsuccessful ExpectationValidationResult
        Args:
            exception_traceback: Traceback related to raised Exception
            exception: Exception raised
            failing_expectation_configurations: ExpectationConfigurations that failed
            evrs: List of ExpectationValidationResult objects to append failures to

        Returns:
            List of ExpectationValidationResult objects with unsuccessful ExpectationValidationResult objects appended
        """  # noqa: E501 # FIXME CoP
        exception_message: str = str(exception)
        exception_info = ExceptionInfo(
            exception_traceback=exception_traceback,
            exception_message=exception_message,
        )

        configuration: ExpectationConfiguration
        result: ExpectationValidationResult
        for configuration in failing_expectation_configurations:
            result = ExpectationValidationResult(
                success=False,
                exception_info=exception_info,
                expectation_config=configuration,
            )
            evrs.append(result)

        return evrs

    def remove_expectation(
        self,
        expectation_configuration: ExpectationConfiguration,
        match_type: str = "domain",
        remove_multiple_matches: bool = False,
        id: Optional[str] = None,
    ) -> List[ExpectationConfiguration]:
        """Remove an ExpectationConfiguration from the ExpectationSuite associated with the Validator.

        Args:
            expectation_configuration: A potentially incomplete (partial) Expectation Configuration to match against.
            match_type: This determines what kwargs to use when matching. Options are:
                - 'domain' to match based on the data evaluated by that expectation
                - 'success' to match based on all configuration parameters that influence whether an expectation succeeds on a given batch of data
                - 'runtime' to match based on all configuration parameters.
            remove_multiple_matches: If True, will remove multiple matching expectations.
            id: Great Expectations Cloud id for an Expectation.

        Returns:
            The list of deleted ExpectationConfigurations.

        Raises:
            TypeError: Must provide either expectation_configuration or id.
            ValueError: No match or multiple matches found (and remove_multiple_matches=False).
        """  # noqa: E501 # FIXME CoP

        return self._expectation_suite.remove_expectation(
            expectation_configuration=expectation_configuration,
            match_type=match_type,
            remove_multiple_matches=remove_multiple_matches,
            id=id,
        )

    def discard_failing_expectations(self) -> None:
        """Removes any expectations from the validator where the validation has failed"""
        res = self.validate(only_return_failures=True).results  # type: ignore[union-attr] # ExpectationValidationResult has no `.results` attr
        if any(res):
            for item in res:
                config = item.expectation_config
                if not config:
                    raise ValueError(  # noqa: TRY003 # FIXME CoP
                        "ExpectationValidationResult does not have an expectation_config"
                    )
                self.remove_expectation(
                    expectation_configuration=config,
                    match_type="runtime",
                )
            warnings.warn(f"Removed {len(res)} expectations that were 'False'")

    def get_default_expectation_arguments(self) -> dict:
        """Fetch default expectation arguments for this data_asset

        Returns:
            A dictionary containing all the current default expectation arguments for a data_asset

            Ex::

                {
                    "catch_exceptions" : False,
                    "result_format" : 'BASIC'
                }

        See also:
            set_default_expectation_arguments
        """
        return self.default_expectation_args

    @property
    def cloud_mode(self) -> bool:
        """
        Wrapper around cloud_mode property of associated Data Context
        """
        from great_expectations.data_context.data_context.cloud_data_context import (
            CloudDataContext,
        )

        return isinstance(self._data_context, CloudDataContext)

    @property
    def ge_cloud_mode(self) -> bool:
        # <GE_RENAME> Deprecated 0.15.37
        return self.cloud_mode

    @property
    def default_expectation_args(self) -> dict:
        """A getter for default Expectation arguments"""
        return self._default_expectation_args

    def set_default_expectation_argument(self, argument: str, value) -> None:
        """
        Set a default expectation argument for this data_asset

        Args:
            argument (string): The argument to be replaced
            value : The New argument to use for replacement

        Returns:
            None

        See also:
            get_default_expectation_arguments
        """

        self._default_expectation_args[argument] = value

    def get_expectation_suite(  # noqa: C901, PLR0912, PLR0913 # FIXME CoP
        self,
        discard_failed_expectations: bool = True,
        discard_result_format_kwargs: bool = True,
        discard_include_config_kwargs: bool = True,
        discard_catch_exceptions_kwargs: bool = True,
        suppress_warnings: bool = False,
        suppress_logging: bool = False,
    ) -> ExpectationSuite:
        """Get a copy of the Expectation Suite from the Validator object.

        Args:
            discard_failed_expectations: Omit Expectations which failed on their last run.
            discard_result_format_kwargs: Omit `result_format` from each Expectation.
            discard_include_config_kwargs: Omit `include_config` from each Expectation.
            discard_catch_exceptions_kwargs: Omit `catch_exceptions` from each Expectation.
            suppress_warnings: Do not log warnings.
            suppress_logging: Do not log anything.

        Returns:
            ExpectationSuite object.
        """

        expectation_suite = copy.deepcopy(self.expectation_suite)
        expectations = expectation_suite.expectation_configurations

        discards: defaultdict[str, int] = defaultdict(int)

        if discard_failed_expectations:
            new_expectations = []

            for expectation in expectations:
                # Note: This is conservative logic.
                # Instead of retaining expectations IFF success==True, it discard expectations IFF success==False.  # noqa: E501 # FIXME CoP
                # In cases where expectation.success is missing or None, expectations are *retained*.  # noqa: E501 # FIXME CoP
                # Such a case could occur if expectations were loaded from a config file and never run.  # noqa: E501 # FIXME CoP
                if expectation.success_on_last_run is False:
                    discards["failed_expectations"] += 1
                else:
                    new_expectations.append(expectation)

            expectations = new_expectations

        message = f"\t{len(expectations)} expectation(s) included in expectation_suite."

        if discards["failed_expectations"] > 0 and not suppress_warnings:
            message += (
                f" Omitting {discards['failed_expectations']} expectation(s) that failed when last run; set "  # noqa: E501 # FIXME CoP
                "discard_failed_expectations=False to include them."
            )

        for expectation in expectations:
            # FIXME: Factor this out into a new function. The logic is duplicated in remove_expectation,  # noqa: E501 # FIXME CoP
            #  which calls _copy_and_clean_up_expectation
            expectation.success_on_last_run = None

            if discard_result_format_kwargs:
                if "result_format" in expectation.kwargs:
                    del expectation.kwargs["result_format"]
                    discards["result_format"] += 1

            if discard_include_config_kwargs:
                if "include_config" in expectation.kwargs:
                    del expectation.kwargs["include_config"]
                    discards["include_config"] += 1

            if discard_catch_exceptions_kwargs:
                if "catch_exceptions" in expectation.kwargs:
                    del expectation.kwargs["catch_exceptions"]
                    discards["catch_exceptions"] += 1

        settings_message = ""

        if discards["result_format"] > 0 and not suppress_warnings:
            settings_message += " result_format"

        if discards["include_config"] > 0 and not suppress_warnings:
            settings_message += " include_config"

        if discards["catch_exceptions"] > 0 and not suppress_warnings:
            settings_message += " catch_exceptions"

        if len(settings_message) > 1:  # Only add this if we added one of the settings above.
            settings_message += " settings filtered."

        expectation_suite.expectations = []
        expectation_suite.add_expectation_configurations(expectation_configurations=expectations)
        if not suppress_logging:
            logger.info(message + settings_message)
        return expectation_suite

    def save_expectation_suite(  # noqa: PLR0913 # FIXME CoP
        self,
        filepath: Optional[str] = None,
        discard_failed_expectations: bool = True,
        discard_result_format_kwargs: bool = True,
        discard_include_config_kwargs: bool = True,
        discard_catch_exceptions_kwargs: bool = True,
        suppress_warnings: bool = False,
    ) -> None:
        """Write the Expectation Suite (e.g. from interactive evaluation) to the Expectation Store associated with the Validator's Data Context.

        If `filepath` is provided, the Data Context configuration will be ignored and the configuration will be written, as JSON, to the specified file.

        Args:
            filepath: The location and name to write the JSON config file to. This parameter overrides the Data Context configuration.
            discard_failed_expectations: If True, excludes expectations that do not return `success = True`. If False, all expectations are saved.
            discard_result_format_kwargs: If True, the `result_format` attribute for each expectation is not included in the saved configuration.
            discard_include_config_kwargs: If True, the `include_config` attribute for each expectation is not included in the saved configuration.
            discard_catch_exceptions_kwargs: If True, the `catch_exceptions` attribute for each expectation is not included in the saved configuration.
            suppress_warnings: If True, all warnings raised by Great Expectations, as a result of dropped expectations, are suppressed.

        Raises:
            ValueError: Must configure a Data Context when instantiating the Validator or pass in `filepath`.
        """  # noqa: E501 # FIXME CoP
        expectation_suite: ExpectationSuite = self.get_expectation_suite(
            discard_failed_expectations,
            discard_result_format_kwargs,
            discard_include_config_kwargs,
            discard_catch_exceptions_kwargs,
            suppress_warnings,
        )
        if filepath is None and self._data_context is not None:
            self._data_context.suites.add(expectation_suite)
            if self.cloud_mode:
                updated_suite = self._data_context.suites.get(expectation_suite.name)
                self._initialize_expectations(expectation_suite=updated_suite)
        elif filepath is not None:
            with open(filepath, "w") as outfile:
                json.dump(
                    expectationSuiteSchema.dump(expectation_suite),
                    outfile,
                    indent=2,
                    sort_keys=True,
                )
        else:
            raise ValueError("Unable to save config: filepath or data_context must be available.")  # noqa: TRY003 # FIXME CoP

    @deprecated_argument(
        argument_name="run_id",
        message="Only the str version of this argument is deprecated. run_id should be a RunIdentifier or dict. Support will be removed in 0.16.0.",  # noqa: E501 # FIXME CoP
        version="0.13.0",
    )
    def validate(  # noqa: C901, PLR0912, PLR0913 # FIXME CoP
        self,
        expectation_suite: str | ExpectationSuite | None = None,
        run_id: str | RunIdentifier | Dict[str, str] | None = None,
        data_context: Optional[Any] = None,  # Cannot type DataContext due to circular import
        suite_parameters: Optional[dict] = None,
        catch_exceptions: bool = True,
        result_format: Optional[str] = None,
        only_return_failures: bool = False,
        run_name: Optional[str] = None,
        run_time: Optional[str] = None,
        checkpoint_name: Optional[str] = None,
    ) -> Union[ExpectationValidationResult, ExpectationSuiteValidationResult]:
        # noinspection SpellCheckingInspection
        """Run all expectations and return the outcome of the run.

        Args:
            expectation_suite: If None, uses the Expectation Suite configuration generated during the current Validator session. If an `ExpectationSuite` object, uses it as the configuration. If a string, assumes it is a path to a JSON file, and loads it as the Expectation Sutie configuration.
            run_id: Used to identify this validation result as part of a collection of validations.
            run_name: Used to identify this validation result as part of a collection of validations. Only used if a `run_id` is not passed. See DataContext for more information.
            run_time: Used to identify this validation result as part of a collection of validations. Only used if a `run_id` is not passed. See DataContext for more information.
            data_context: A datacontext object to use as part of validation for binding suite parameters and registering validation results. Overrides the Data Context configured when the Validator is instantiated.
            suite_parameters: If None, uses the evaluation_paramters from the Expectation Suite provided or as part of the Data Asset. If a dict, uses the suite parameters in the dictionary.
            catch_exceptions: If True, exceptions raised by tests will not end validation and will be described in the returned report.
            result_format: If None, uses the default value ('BASIC' or as specified). If string, the returned expectation output follows the specified format ('BOOLEAN_ONLY','BASIC', etc.).
            only_return_failures: If True, expectation results are only returned when `success = False`.
            checkpoint_name: Name of the Checkpoint which invoked this Validator.validate() call against an Expectation Suite. It will be added to `meta` field of the returned ExpectationSuiteValidationResult.

        Returns:
            Object containg the results.

        Raises:
            Exception: Depending on the Data Context configuration and arguments, there are numerous possible exceptions that may be raised.
            GreatExpectationsError: If `expectation_suite` is a string it must point to an existing and readable file.
            ValidationError: If `expectation_suite` is a string, the file it points to must be valid JSON.

        """  # noqa: E501 # FIXME CoP
        # noinspection PyUnusedLocal
        try:
            validation_time = datetime.datetime.now(datetime.timezone.utc).strftime(
                "%Y%m%dT%H%M%S.%fZ"
            )
            assert not (run_id and run_name) and not (run_id and run_time), (
                "Please provide either a run_id or run_name and/or run_time."
            )
            if isinstance(run_id, dict):
                run_id = RunIdentifier(**run_id)
            elif not isinstance(run_id, RunIdentifier):
                run_id = RunIdentifier(run_name=run_name, run_time=run_time)

            self._active_validation = True

            # If a different validation data context was provided, override
            validation_data_context = self._data_context
            if data_context is None and self._data_context is not None:
                data_context = self._data_context
            elif data_context is not None:
                # temporarily set self._data_context so it is used inside the expectation decorator
                self._data_context = data_context

            if expectation_suite is None:
                expectation_suite = self.get_expectation_suite(
                    discard_failed_expectations=False,
                    discard_result_format_kwargs=False,
                    discard_include_config_kwargs=False,
                    discard_catch_exceptions_kwargs=False,
                )
            elif isinstance(expectation_suite, str):
                try:
                    with open(expectation_suite) as infile:
                        expectation_suite = expectationSuiteSchema.loads(infile.read())
                except ValidationError:
                    raise
                except OSError:
                    raise GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                        f"Unable to load expectation suite: IO error while reading {expectation_suite}"  # noqa: E501 # FIXME CoP
                    )

            if not isinstance(expectation_suite, ExpectationSuite):
                logger.error(
                    "Unable to validate using the provided value for expectation suite; does it need to be "  # noqa: E501 # FIXME CoP
                    "loaded from a dictionary?"
                )
                return ExpectationValidationResult(success=False)

            # Suite parameter priority is
            # 1. from provided parameters
            # 2. from expectation configuration
            # 3. from data context
            # So, we load them in reverse order

            runtime_suite_parameters: dict = {}

            if expectation_suite.suite_parameters:
                runtime_suite_parameters.update(expectation_suite.suite_parameters)

            if suite_parameters is not None:
                runtime_suite_parameters.update(suite_parameters)

            # Convert suite parameters to be json-serializable
            runtime_suite_parameters = recursively_convert_to_json_serializable(
                runtime_suite_parameters
            )

            # Warn if our version is different from the version in the configuration
            # TODO: Deprecate "great_expectations.__version__"

            expectations_to_evaluate = self.process_expectations_for_validation(
                expectation_suite.expectation_configurations,
                runtime_suite_parameters,
            )

            runtime_configuration = self._get_runtime_configuration(
                catch_exceptions=catch_exceptions, result_format=result_format
            )

            results = self.graph_validate(
                configurations=expectations_to_evaluate,
                runtime_configuration=runtime_configuration,
            )

            if self._include_rendered_content:
                for validation_result in results:
                    validation_result.render()

            statistics = calc_validation_statistics(results)

            if only_return_failures:
                abbrev_results = []
                for exp in results:
                    if not exp.success:
                        abbrev_results.append(exp)
                results = abbrev_results

            expectation_suite_name = expectation_suite.name

            result = ExpectationSuiteValidationResult(
                results=results,
                success=statistics.success,
                suite_name=expectation_suite_name,
                statistics={
                    "evaluated_expectations": statistics.evaluated_expectations,
                    "successful_expectations": statistics.successful_expectations,
                    "unsuccessful_expectations": statistics.unsuccessful_expectations,
                    "success_percent": statistics.success_percent,
                },
                suite_parameters=runtime_suite_parameters,
                meta={
                    "great_expectations_version": ge_version,
                    "expectation_suite_name": expectation_suite_name,
                    "run_id": run_id,
                    "batch_spec": convert_to_json_serializable(self.active_batch_spec),
                    "batch_markers": self.active_batch_markers,
                    "active_batch_definition": self.active_batch_definition,
                    "validation_time": validation_time,
                    "checkpoint_name": checkpoint_name,
                },
                batch_id=self.active_batch_id,
            )

            self._data_context = validation_data_context
        finally:
            self._active_validation = False

        return result

    def process_expectations_for_validation(
        self,
        expectation_configurations: list[ExpectationConfiguration],
        suite_parameters: Optional[dict[str, Any]] = None,
    ) -> list[ExpectationConfiguration]:
        """Substitute suite parameters into the provided expectations and sort by column."""
        NO_COLUMN = "_nocolumn"  # just used to group expectations that don't specify a column
        columns: dict[str, list[ExpectationConfiguration]] = {}

        for expectation in expectation_configurations:
            expectation.process_suite_parameters(
                suite_parameters=suite_parameters,
                interactive_evaluation=self.interactive_evaluation,
                data_context=self._data_context,
            )
            if "column" in expectation.kwargs and isinstance(expectation.kwargs["column"], str):
                column = expectation.kwargs["column"]
            else:
                column = NO_COLUMN
            if column not in columns:
                columns[column] = []
            columns[column].append(expectation)

        expectations_to_evaluate = []
        for col in columns.values():
            expectations_to_evaluate.extend(col)

        return expectations_to_evaluate

    def get_suite_parameter(self, parameter_name, default_value=None):
        """
        Get an suite parameter value that has been stored in meta.

        Args:
            parameter_name (string): The name of the parameter to store.
            default_value (any): The default value to be returned if the parameter is not found.

        Returns:
            The current value of the suite parameter.
        """
        if parameter_name in self._expectation_suite.suite_parameters:
            return self._expectation_suite.suite_parameters[parameter_name]
        else:
            return default_value

    def set_suite_parameter(self, parameter_name, parameter_value) -> None:
        """
        Provide a value to be stored in the data_asset suite_parameters object and used to evaluate
        parameterized expectations.

        Args:
            parameter_name (string): The name of the kwarg to be replaced at evaluation time
            parameter_value (any): The value to be used
        """
        self._expectation_suite.suite_parameters.update(
            {parameter_name: convert_to_json_serializable(parameter_value)}
        )

    def test_expectation_function(self, function: Callable, *args, **kwargs) -> Callable:
        """Test a generic expectation function

        Args:
            function (func): The function to be tested. (Must be a valid expectation function.)
            *args          : Positional arguments to be passed the function
            **kwargs       : Keyword arguments to be passed the function

        Returns:
            A JSON-serializable expectation result object.

        Notes:
            This function is a thin layer to allow quick testing of new expectation functions, without having to \
            define custom classes, etc. To use developed expectations from the command-line tool, you will still need \
            to define custom classes, etc.

            Check out :ref:`how_to_guides__creating_and_editing_expectations__how_to_create_custom_expectations` for
            more information.
        """  # noqa: E501 # FIXME CoP

        # noinspection SpellCheckingInspection
        argspec = inspect.getfullargspec(function)[0][1:]

        new_function = self.expectation(argspec)(function)
        return new_function(self, *args, **kwargs)

    @staticmethod
    def _parse_validation_graph(
        validation_graph: ValidationGraph,
        metrics: _MetricsDict,
    ) -> Tuple[Set[MetricConfiguration], Set[MetricConfiguration]]:
        """Given validation graph, returns the ready and needed metrics necessary for validation using a traversal of
        validation graph (a graph structure of metric ids) edges"""  # noqa: E501 # FIXME CoP
        unmet_dependency_ids = set()
        unmet_dependency = set()
        maybe_ready_ids = set()
        maybe_ready = set()

        for edge in validation_graph.edges:
            if edge.left.id not in metrics:
                if edge.right is None or edge.right.id in metrics:
                    if edge.left.id not in maybe_ready_ids:
                        maybe_ready_ids.add(edge.left.id)
                        maybe_ready.add(edge.left)
                else:  # noqa: PLR5501 # FIXME CoP
                    if edge.left.id not in unmet_dependency_ids:
                        unmet_dependency_ids.add(edge.left.id)
                        unmet_dependency.add(edge.left)

        return maybe_ready - unmet_dependency, unmet_dependency

    def _initialize_expectations(
        self,
        expectation_suite: Optional[ExpectationSuite] = None,
        expectation_suite_name: Optional[str] = None,
    ) -> None:
        """Instantiates `_expectation_suite` as empty by default or with a specified expectation `config`.
        In addition, this always sets the `default_expectation_args` to:
            `include_config`: False,
            `catch_exceptions`: False,
            `output_format`: 'BASIC'

        By default, initializes data_asset_type to the name of the implementing class, but subclasses
        that have interoperable semantics (e.g. Dataset) may override that parameter to clarify their
        interoperability.

        Args:
            expectation_suite (json): \
                A json-serializable expectation config. \
                If None, creates default `_expectation_suite` with an empty list of expectations and \
                key value `data_asset_name` as `data_asset_name`.

            expectation_suite_name (string): \
                The name to assign to the `expectation_suite.name`

        Returns:
            None
        """  # noqa: E501 # FIXME CoP
        # Checking type of expectation_suite.
        # Check for expectation_suite_name is already done by ExpectationSuiteIdentifier
        if expectation_suite and not isinstance(expectation_suite, ExpectationSuite):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"expectation_suite must be of type ExpectationSuite, not {type(expectation_suite)}"
            )
        if expectation_suite is not None:
            if isinstance(expectation_suite, dict):
                expectation_suite_dict: dict = expectationSuiteSchema.load(expectation_suite)
                expectation_suite = ExpectationSuite(**expectation_suite_dict)
            else:
                expectation_suite = copy.deepcopy(expectation_suite)
            self._expectation_suite: ExpectationSuite = expectation_suite

            if expectation_suite_name is not None:
                if self._expectation_suite.name != expectation_suite_name:
                    logger.warning(
                        f"Overriding existing expectation_suite_name {self._expectation_suite.name} with new name {expectation_suite_name}"  # noqa: E501 # FIXME CoP
                    )
                self._expectation_suite.name = expectation_suite_name

        else:
            if expectation_suite_name is None:
                expectation_suite_name = "default"
            self._expectation_suite = ExpectationSuite(name=expectation_suite_name)

    def _get_runtime_configuration(
        self,
        catch_exceptions: Optional[bool] = None,
        result_format: Optional[Union[dict, str]] = None,
    ) -> dict:
        runtime_configuration = copy.deepcopy(self.default_expectation_args)

        if catch_exceptions is not None:
            runtime_configuration.update({"catch_exceptions": catch_exceptions})

        if (
            self.default_expectation_args["result_format"]
            == Validator.DEFAULT_RUNTIME_CONFIGURATION["result_format"]
        ):
            if result_format is None:
                runtime_configuration.pop("result_format")
            else:
                runtime_configuration.update({"result_format": result_format})
        else:  # noqa: PLR5501 # FIXME CoP
            if result_format is not None:
                runtime_configuration.update({"result_format": result_format})

        return runtime_configuration

    def convert_to_checkpoint_validations_list(
        self,
    ) -> list[CheckpointValidationDefinition]:
        """
        Generates a list of validations to be used in the construction of a Checkpoint.

        Returns:
            A list of CheckpointValidationDefinitions (one for each batch in the Validator).
        """
        validations = []
        for batch in self.batch_cache.values():
            validation = CheckpointValidationDefinition(
                expectation_suite_name=self.expectation_suite_name,
                expectation_suite_id=self.expectation_suite.id,
                batch_request=batch.batch_request,
            )
            validations.append(validation)

        return validations



from __future__ import annotations

from great_expectations.core.batch import BatchData


class SparkDFBatchData(BatchData):
    def __init__(self, execution_engine, dataframe) -> None:
        super().__init__(execution_engine=execution_engine)
        self._dataframe = dataframe

    @property
    def dataframe(self):
        return self._dataframe



from __future__ import annotations

from great_expectations.compatibility.pyspark import functions as F
from great_expectations.compatibility.pyspark import types
from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.column_aggregate_metric_provider import (
    ColumnAggregateMetricProvider,
    column_aggregate_partial,
    column_aggregate_value,
)
from great_expectations.util import convert_pandas_series_decimal_to_float_dtype


class ColumnMean(ColumnAggregateMetricProvider):
    """MetricProvider Class for Aggregate Mean MetricProvider"""

    metric_name = "column.mean"

    @column_aggregate_value(engine=PandasExecutionEngine)
    def _pandas(cls, column, **kwargs):
        """Pandas Mean Implementation"""
        convert_pandas_series_decimal_to_float_dtype(data=column, inplace=True)
        return column.mean()

    @column_aggregate_partial(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(cls, column, **kwargs):
        """SqlAlchemy Mean Implementation"""
        # column * 1.0 needed for correct calculation of avg in MSSQL
        return sa.func.avg(1.0 * column)

    @column_aggregate_partial(engine=SparkDFExecutionEngine)
    def _spark(cls, column, _table, _column_name, **kwargs):
        """Spark Mean Implementation"""
        column_data_type = _table.schema[_column_name].dataType
        if type(column_data_type) not in (
            types.DecimalType,
            types.IntegerType,
            types.DoubleType,
            types.FloatType,
            types.LongType,
        ):
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"Expected numeric column type for function mean(). Recieved type: {column_data_type}"  # noqa: E501 # FIXME CoP
            )
        return F.mean(column)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, ClassVar, Literal, Type

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.util import DBFSPath
from great_expectations.datasource.fluent import PandasFilesystemDatasource
from great_expectations.datasource.fluent.data_connector import (
    DBFSDataConnector,
)

if TYPE_CHECKING:
    from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset

logger = logging.getLogger(__name__)


@public_api
class PandasDBFSDatasource(PandasFilesystemDatasource):
    """Pandas based Datasource for DataBricks File System (DBFS) based data assets."""

    # class attributes
    data_connector_type: ClassVar[Type[DBFSDataConnector]] = DBFSDataConnector

    # instance attributes
    # overridden from base `Literal['pandas_filesystem']`
    type: Literal["pandas_dbfs"] = "pandas_dbfs"  # type: ignore[assignment] # base class has different type

    @override
    def _build_data_connector(
        self, data_asset: FileDataAsset, glob_directive: str = "**/*", **kwargs
    ) -> None:
        """Builds and attaches the `DBFSDataConnector` to the asset."""
        if kwargs:
            raise TypeError(  # noqa: TRY003 # FIXME CoP
                f"_build_data_connector() got unexpected keyword arguments {list(kwargs.keys())}"
            )
        data_asset._data_connector = self.data_connector_type.build_data_connector(
            datasource_name=self.name,
            data_asset_name=data_asset.name,
            base_directory=self.base_directory,
            glob_directive=glob_directive,
            data_context_root_directory=self.data_context_root_directory,
            file_path_template_map_fn=DBFSPath.convert_to_file_semantics_version,
        )

        # build a more specific `_test_connection_error_message`
        data_asset._test_connection_error_message = (
            self.data_connector_type.build_test_connection_error_message(
                data_asset_name=data_asset.name,
                glob_directive=glob_directive,
                base_directory=self.base_directory,
            )
        )



import typing
from logging import Logger
from typing import Hashable, Iterable, Literal, Optional, Sequence, Union

from great_expectations._docs_decorators import public_api as public_api
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent import PandasFilesystemDatasource
from great_expectations.datasource.fluent.data_asset.path.pandas.generated_assets import (
    CSVAsset,
    ExcelAsset,
    FeatherAsset,
    HDFAsset,
    HTMLAsset,
    JSONAsset,
    ORCAsset,
    ParquetAsset,
    PickleAsset,
    SASAsset,
    SPSSAsset,
    StataAsset,
    XMLAsset,
)
from great_expectations.datasource.fluent.data_connector import (
    DBFSDataConnector as DBFSDataConnector,
)
from great_expectations.datasource.fluent.dynamic_pandas import (
    CompressionOptions,
    CSVEngine,
    FilePath,
    IndexLabel,
    StorageOptions,
)
from great_expectations.datasource.fluent.interfaces import BatchMetadata
from great_expectations.datasource.fluent.interfaces import (
    SortersDefinition as SortersDefinition,
)
from great_expectations.datasource.fluent.interfaces import (
    TestConnectionError as TestConnectionError,
)

logger: Logger

class PandasDBFSDatasource(PandasFilesystemDatasource):
    type: Literal["pandas_dbfs"]  # type: ignore[assignment] # FIXME CoP

    @override
    def add_csv_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        sep: typing.Union[str, None] = ...,
        delimiter: typing.Union[str, None] = ...,
        header: Union[int, Sequence[int], None, Literal["infer"]] = "infer",
        names: Union[Sequence[Hashable], None] = ...,
        index_col: Union[IndexLabel, Literal[False], None] = ...,
        usecols: typing.Union[int, str, typing.Sequence[int], None] = ...,
        squeeze: typing.Union[bool, None] = ...,
        prefix: str = ...,
        mangle_dupe_cols: bool = ...,
        dtype: typing.Union[dict, None] = ...,
        engine: Union[CSVEngine, None] = ...,
        converters: typing.Any = ...,
        true_values: typing.Any = ...,
        false_values: typing.Any = ...,
        skipinitialspace: bool = ...,
        skiprows: typing.Union[typing.Sequence[int], int, None] = ...,
        skipfooter: int = 0,
        nrows: typing.Union[int, None] = ...,
        na_values: typing.Any = ...,
        keep_default_na: bool = ...,
        na_filter: bool = ...,
        verbose: bool = ...,
        skip_blank_lines: bool = ...,
        parse_dates: typing.Any = ...,
        infer_datetime_format: bool = ...,
        keep_date_col: bool = ...,
        date_parser: typing.Any = ...,
        dayfirst: bool = ...,
        cache_dates: bool = ...,
        iterator: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        compression: CompressionOptions = "infer",
        thousands: typing.Union[str, None] = ...,
        decimal: str = ".",
        lineterminator: typing.Union[str, None] = ...,
        quotechar: str = '"',
        quoting: int = 0,
        doublequote: bool = ...,
        escapechar: typing.Union[str, None] = ...,
        comment: typing.Union[str, None] = ...,
        encoding: typing.Union[str, None] = ...,
        encoding_errors: typing.Union[str, None] = "strict",
        dialect: typing.Union[str, None] = ...,
        error_bad_lines: typing.Union[bool, None] = ...,
        warn_bad_lines: typing.Union[bool, None] = ...,
        on_bad_lines: typing.Any = ...,
        delim_whitespace: bool = ...,
        low_memory: typing.Any = ...,
        memory_map: bool = ...,
        storage_options: StorageOptions = ...,
    ) -> CSVAsset: ...
    @override
    def add_excel_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        sheet_name: typing.Union[str, int, None] = 0,
        header: Union[int, Sequence[int], None] = 0,
        names: typing.Union[typing.List[str], None] = ...,
        index_col: Union[int, Sequence[int], None] = ...,
        usecols: typing.Union[int, str, typing.Sequence[int], None] = ...,
        squeeze: typing.Union[bool, None] = ...,
        dtype: typing.Union[dict, None] = ...,
        true_values: Union[Iterable[Hashable], None] = ...,
        false_values: Union[Iterable[Hashable], None] = ...,
        skiprows: typing.Union[typing.Sequence[int], int, None] = ...,
        nrows: typing.Union[int, None] = ...,
        na_values: typing.Any = ...,
        keep_default_na: bool = ...,
        na_filter: bool = ...,
        verbose: bool = ...,
        parse_dates: typing.Union[typing.List, typing.Dict, bool] = ...,
        thousands: typing.Union[str, None] = ...,
        decimal: str = ".",
        comment: typing.Union[str, None] = ...,
        skipfooter: int = 0,
        convert_float: typing.Union[bool, None] = ...,
        mangle_dupe_cols: bool = ...,
        storage_options: StorageOptions = ...,
    ) -> ExcelAsset: ...
    @override
    def add_feather_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        columns: Union[Sequence[Hashable], None] = ...,
        use_threads: bool = ...,
        storage_options: StorageOptions = ...,
    ) -> FeatherAsset: ...
    @override
    def add_hdf_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        key: typing.Any = ...,
        mode: str = "r",
        errors: str = "strict",
        where: typing.Union[str, typing.List, None] = ...,
        start: typing.Union[int, None] = ...,
        stop: typing.Union[int, None] = ...,
        columns: typing.Union[typing.List[str], None] = ...,
        iterator: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        kwargs: typing.Union[dict, None] = ...,
    ) -> HDFAsset: ...
    @override
    def add_html_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        match: Union[str, typing.Pattern] = ".+",
        flavor: typing.Union[str, None] = ...,
        header: Union[int, Sequence[int], None] = ...,
        index_col: Union[int, Sequence[int], None] = ...,
        skiprows: typing.Union[typing.Sequence[int], int, None] = ...,
        attrs: typing.Union[typing.Dict[str, str], None] = ...,
        parse_dates: bool = ...,
        thousands: typing.Union[str, None] = ",",
        encoding: typing.Union[str, None] = ...,
        decimal: str = ".",
        converters: typing.Union[typing.Dict, None] = ...,
        na_values: Union[Iterable[object], None] = ...,
        keep_default_na: bool = ...,
        displayed_only: bool = ...,
    ) -> HTMLAsset: ...
    @override
    def add_json_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        orient: typing.Union[str, None] = ...,
        dtype: typing.Union[dict, None] = ...,
        convert_axes: typing.Any = ...,
        convert_dates: typing.Union[bool, typing.List[str]] = ...,
        keep_default_dates: bool = ...,
        numpy: bool = ...,
        precise_float: bool = ...,
        date_unit: typing.Union[str, None] = ...,
        encoding: typing.Union[str, None] = ...,
        encoding_errors: typing.Union[str, None] = "strict",
        lines: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        compression: CompressionOptions = "infer",
        nrows: typing.Union[int, None] = ...,
        storage_options: StorageOptions = ...,
    ) -> JSONAsset: ...
    @override
    def add_orc_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        columns: typing.Union[typing.List[str], None] = ...,
        kwargs: typing.Union[dict, None] = ...,
    ) -> ORCAsset: ...
    @override
    def add_parquet_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        engine: str = "auto",
        columns: typing.Union[typing.List[str], None] = ...,
        storage_options: StorageOptions = ...,
        use_nullable_dtypes: bool = ...,
        kwargs: typing.Union[dict, None] = ...,
    ) -> ParquetAsset: ...
    @override
    def add_pickle_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        compression: CompressionOptions = "infer",
        storage_options: StorageOptions = ...,
    ) -> PickleAsset: ...
    @override
    def add_sas_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        format: typing.Union[str, None] = ...,
        index: Union[Hashable, None] = ...,
        encoding: typing.Union[str, None] = ...,
        chunksize: typing.Union[int, None] = ...,
        iterator: bool = ...,
        compression: CompressionOptions = "infer",
    ) -> SASAsset: ...
    @override
    def add_spss_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        usecols: typing.Union[int, str, typing.Sequence[int], None] = ...,
        convert_categoricals: bool = ...,
    ) -> SPSSAsset: ...
    @override
    def add_stata_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        convert_dates: bool = ...,
        convert_categoricals: bool = ...,
        index_col: typing.Union[str, None] = ...,
        convert_missing: bool = ...,
        preserve_dtypes: bool = ...,
        columns: Union[Sequence[str], None] = ...,
        order_categoricals: bool = ...,
        chunksize: typing.Union[int, None] = ...,
        iterator: bool = ...,
        compression: CompressionOptions = "infer",
        storage_options: StorageOptions = ...,
    ) -> StataAsset: ...
    @override
    def add_xml_asset(
        self,
        name: str,
        *,
        glob_directive: str = ...,
        batch_metadata: Optional[BatchMetadata] = ...,
        xpath: str = "./*",
        namespaces: typing.Union[typing.Dict[str, str], None] = ...,
        elems_only: bool = ...,
        attrs_only: bool = ...,
        names: Union[Sequence[str], None] = ...,
        dtype: typing.Union[dict, None] = ...,
        encoding: typing.Union[str, None] = "utf-8",
        stylesheet: Union[FilePath, None] = ...,
        iterparse: typing.Union[typing.Dict[str, typing.List[str]], None] = ...,
        compression: CompressionOptions = "infer",
        storage_options: StorageOptions = ...,
    ) -> XMLAsset: ...



from datetime import date, datetime, timezone
from typing import Any

import pytest
import sqlalchemy.sql

from great_expectations.core.batch_definition import BatchDefinition
from great_expectations.core.partitioners import (
    ColumnPartitionerDaily,
    ColumnPartitionerMonthly,
    ColumnPartitionerYearly,
)
from great_expectations.datasource.fluent import SQLDatasource
from great_expectations.datasource.fluent.sql_datasource import (
    SqlAddBatchDefinitionError,
    TableAsset,
    _SQLAsset,
)
from tests.datasource.fluent.conftest import CreateSourceFixture


@pytest.fixture
def postgres_asset(empty_data_context, create_source: CreateSourceFixture, monkeypatch):
    my_config_variables = {"pipeline_filename": __file__}
    empty_data_context.config_variables.update(my_config_variables)

    monkeypatch.setattr(TableAsset, "test_connection", lambda _: None)
    years = [2021, 2022]
    asset_specified_metadata = {
        "pipeline_name": "my_pipeline",
        "no_curly_pipeline_filename": "$pipeline_filename",
        "curly_pipeline_filename": "${pipeline_filename}",
    }

    with create_source(
        validate_batch_spec=lambda _: None,
        dialect="postgresql",
        data_context=empty_data_context,
        partitioner_query_response=[{"year": year} for year in years],
    ) as source:
        asset = source.add_table_asset(
            name="query_asset",
            table_name="my_table",
            batch_metadata=asset_specified_metadata,
        )
        assert asset.batch_metadata == asset_specified_metadata

        yield asset


@pytest.mark.postgresql
def test_get_batch_identifiers_list__sort_ascending(postgres_asset):
    years = [2021, 2022]
    batches = postgres_asset.get_batch_identifiers_list(
        postgres_asset.build_batch_request(
            partitioner=ColumnPartitionerYearly(column_name="year", sort_ascending=True)
        )
    )

    assert len(batches) == len(years)
    for i, year in enumerate([2021, 2022]):
        batches[i]["year"] = year


@pytest.mark.postgresql
def test_get_batch_identifiers_list__sort_descending(postgres_asset):
    years = [2021, 2022]
    batches = postgres_asset.get_batch_identifiers_list(
        postgres_asset.build_batch_request(
            partitioner=ColumnPartitionerYearly(column_name="year", sort_ascending=False)
        )
    )

    assert len(batches) == len(years)
    for i, year in enumerate([2022, 2021]):
        batches[i]["year"] = year


class FakeResult:
    def __init__(self, queried_row: Any):
        self._queried_row = queried_row

    def first(self):
        # Sqlalchemy will return None if there are no rows when first is called.
        # We can mock that behavior by setting the queried value to None.
        queried_row = self._queried_row

        class Result:
            def __getattr__(self, attr: str):
                if queried_row:
                    return queried_row[0]
                return None

        return Result()


class FakeConnection:
    def __init__(self, queried_row: Any):
        self._queried_row = queried_row

    def execute(self, *args, **kwargs):
        return FakeResult(self._queried_row)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb): ...


class FakeEngineError(Exception):
    def __init__(self):
        super().__init__("We expect at most 1 element in the queried row when using FakeError")


class FakeEngine:
    def __init__(self, queried_row: Any):
        if queried_row and len(queried_row) > 1:
            raise FakeEngineError()
        self._queried_row = queried_row

    def connect(self):
        return FakeConnection(self._queried_row)


@pytest.fixture
def datasource(mocker) -> SQLDatasource:
    return mocker.Mock(spec=SQLDatasource)


@pytest.fixture
def datasource_with_mocked_get_engine(datasource) -> SQLDatasource:
    """Returns a mocked sqlalchemy engine from datasource.get_engine()

    We validate that adding batch definition uses a datetime column by querying
    the database. This patches the datasource so validating the batch definition
    will succeed.
    """
    datasource.get_engine = lambda: FakeEngine([date(2024, 10, 3)])
    return datasource


@pytest.fixture
def asset(datasource, monkeypatch) -> _SQLAsset:
    asset = _SQLAsset[SQLDatasource](name="test_asset", type="_sql_asset")
    asset._datasource = datasource  # same pattern Datasource uses to init Asset

    # _SQLAsset is abstract so we patch in some no-op implementations
    monkeypatch.setattr(
        _SQLAsset[SQLDatasource], "as_selectable", lambda _: sqlalchemy.sql.table("table")
    )

    return asset


@pytest.mark.unit
def test_add_batch_definition_fluent_sql__add_batch_definition_whole_table(
    datasource_with_mocked_get_engine, asset
):
    # arrange
    name = "batch_def_name"
    expected_batch_definition = BatchDefinition(name=name, partitioner=None, batching_regex=None)
    datasource = datasource_with_mocked_get_engine
    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_whole_table(name=name)

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)


@pytest.mark.unit
@pytest.mark.parametrize("sort_ascending", (True, False))
def test_add_batch_definition_fluent_sql__add_batch_definition_yearly(
    datasource_with_mocked_get_engine, asset, sort_ascending
):
    # arrange
    name = "batch_def_name"
    column = "test_column"
    expected_batch_definition = BatchDefinition(
        name=name,
        partitioner=ColumnPartitionerYearly(column_name=column, sort_ascending=sort_ascending),
        batching_regex=None,
    )
    datasource = datasource_with_mocked_get_engine
    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_yearly(
        name=name, column=column, sort_ascending=sort_ascending
    )

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)


@pytest.mark.unit
@pytest.mark.parametrize("sort_ascending", (True, False))
def test_add_batch_definition_fluent_sql__add_batch_definition_monthly(
    datasource_with_mocked_get_engine, asset, sort_ascending
):
    # arrange
    name = "batch_def_name"
    column = "test_column"
    expected_batch_definition = BatchDefinition(
        name=name,
        partitioner=ColumnPartitionerMonthly(column_name=column, sort_ascending=sort_ascending),
        batching_regex=None,
    )
    datasource = datasource_with_mocked_get_engine
    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_monthly(
        name=name, column=column, sort_ascending=sort_ascending
    )

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)


@pytest.mark.unit
@pytest.mark.parametrize("sort_ascending", (True, False))
def test_add_batch_definition_fluent_sql__add_batch_definition_daily(
    datasource_with_mocked_get_engine, asset, sort_ascending
):
    # arrange
    name = "batch_def_name"
    column = "test_column"
    expected_batch_definition = BatchDefinition(
        name=name,
        partitioner=ColumnPartitionerDaily(column_name=column, sort_ascending=sort_ascending),
        batching_regex=None,
    )
    datasource = datasource_with_mocked_get_engine
    datasource.add_batch_definition.return_value = expected_batch_definition

    # act
    batch_definition = asset.add_batch_definition_daily(
        name=name, column=column, sort_ascending=sort_ascending
    )

    # assert
    assert batch_definition == expected_batch_definition
    datasource.add_batch_definition.assert_called_once_with(expected_batch_definition)


def add_table_asset(datasource: SQLDatasource):
    return datasource.add_table_asset(
        name="my_table_asset",
        table_name="my_table",
        schema_name="my_schema",
    )


def add_query_asset(datasource: SQLDatasource):
    return datasource.add_query_asset(name="my_query_asset", query="select * from my_table")


@pytest.mark.unit
@pytest.mark.parametrize(
    "queried_row",
    [
        pytest.param(
            ["not a datetime"],
            id="invalid type",
        ),
        pytest.param([None], id="None row returned"),
        pytest.param(None, id="No rows returned"),
        pytest.param([], id="Empty row returned"),
    ],
)
@pytest.mark.parametrize(
    "add_sql_asset",
    [
        pytest.param(add_table_asset, id="table asset"),
        pytest.param(add_query_asset, id="query asset"),
    ],
)
def test_validate_batch_definition_with_error(
    sql_datasource_table_asset_test_connection_noop: SQLDatasource,
    queried_row: Any,
    add_sql_asset,
    monkeypatch,
):
    # Setup
    # Monkeypatch SQLDatasource so we return a fake engine which returns a fixed db result.
    monkeypatch.setattr(
        SQLDatasource,
        "get_engine",
        lambda _: FakeEngine(queried_row),
    )
    datasource = sql_datasource_table_asset_test_connection_noop

    # Add our asset to the datasource
    asset = add_sql_asset(datasource)

    # Act and assert: Add a batch definition and assert error
    with pytest.raises(SqlAddBatchDefinitionError):
        asset.validate_batch_definition(ColumnPartitionerDaily(column_name="column_name"))


@pytest.mark.unit
@pytest.mark.parametrize(
    "queried_row",
    [
        pytest.param(
            [datetime(2024, 10, 28, 0, 0, 0, tzinfo=timezone.utc)],
            id="datetime",
        ),
        pytest.param([date(2024, 10, 28)], id="date"),
    ],
)
@pytest.mark.parametrize(
    "add_sql_asset",
    [
        pytest.param(add_table_asset, id="table asset"),
        pytest.param(add_query_asset, id="query asset"),
    ],
)
def test_validate_batch_definition(
    sql_datasource_table_asset_test_connection_noop: SQLDatasource,
    queried_row: Any,
    add_sql_asset,
    monkeypatch,
):
    # Setup
    # Monkeypatch SQLDatasource so we return a fake engine which returns a fixed db result.
    monkeypatch.setattr(
        SQLDatasource,
        "get_engine",
        lambda _: FakeEngine(queried_row),
    )
    datasource = sql_datasource_table_asset_test_connection_noop

    # Add our asset to the datasource
    asset = add_sql_asset(datasource)

    # Act
    NoneOnSuccess = asset.validate_batch_definition(
        ColumnPartitionerDaily(column_name="column_name")
    )

    # Assert
    assert NoneOnSuccess is None


# Tests I considered adding for test_validate_batch_definition but have not.
# 1. Engine dies on connect
# 2. Connection dies on execute



from __future__ import annotations

import inspect
import logging
import uuid
import warnings
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    Generator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from great_expectations._docs_decorators import public_api
from great_expectations.compatibility.typing_extensions import override
from great_expectations.datasource.fluent.constants import (
    DEFAULT_PANDAS_DATA_ASSET_NAME,
    DEFAULT_PANDAS_DATASOURCE_NAME,
)
from great_expectations.datasource.fluent.signatures import _merge_signatures
from great_expectations.datasource.fluent.type_lookup import TypeLookup

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from great_expectations.compatibility import pydantic
    from great_expectations.data_context import AbstractDataContext as GXDataContext
    from great_expectations.datasource.datasource_dict import DatasourceDict
    from great_expectations.datasource.fluent import PandasDatasource
    from great_expectations.datasource.fluent.interfaces import DataAsset, Datasource
    from great_expectations.validator.validator import Validator

SourceFactoryFn: TypeAlias = Callable[..., "Datasource"]

logger = logging.getLogger(__name__)


class DefaultPandasDatasourceError(Exception):
    pass


class TypeRegistrationError(TypeError):
    pass


class _FieldDetails(NamedTuple):
    default_value: Any
    type_annotation: Type


def _get_field_details(model: Type[pydantic.BaseModel], field_name: str) -> _FieldDetails:
    """Get the default value of the requested field and its type annotation."""
    return _FieldDetails(
        default_value=model.__fields__[field_name].default,
        type_annotation=model.__fields__[field_name].type_,
    )


class CrudMethodType(str, Enum):
    ADD = "ADD"
    DELETE = "DELETE"  # Deprecated as we don't care about backend-specific deletion
    UPDATE = "UPDATE"
    ADD_OR_UPDATE = "ADD_OR_UPDATE"


CrudMethodInfoFn: TypeAlias = Callable[..., Tuple[CrudMethodType, Type["Datasource"]]]


@public_api
class DataSourceManager:
    """
    Contains methods to interact with data sources from the gx context

    This contains a collection of dynamically generated datasource factory methods in the format
    `.add_<TYPE_NAME>()`.

    It also contains general data source manipulation methods such as `all()`, `get()` and
    `delete()`.
    """

    # A dict-like two way mapping between previously registered `Datasource` or `DataAsset` types
    # and a simplified name for those types.
    type_lookup: ClassVar = TypeLookup()
    __crud_registry: ClassVar[Dict[str, CrudMethodInfoFn]] = {}

    _data_context: GXDataContext

    def __init__(self, data_context: GXDataContext):
        self._data_context = data_context

    @classmethod
    def register_datasource(cls, ds_type: Type[Datasource]) -> None:
        """
        Add/Register a datasource. This registers all the crud datasource methods:
        add_<datasource>, delete_<datasource>, update_<datasource>, add_or_update_<datasource>
        and all related `Datasource`, `DataAsset` and `ExecutionEngine` types.

        Creates mapping table between the `DataSource`/`DataAsset` classes and their
        declared `type` string.


        Example
        -------

        An `.add_pandas_filesystem()` pandas_filesystem factory method will be added to `context.data_sources`.

        >>> class PandasFilesystemDatasource(_PandasFilePathDatasource):
        >>>     type: str = 'pandas_filesystem'
        >>>     asset_types = [FileAsset]
        >>>     execution_engine: PandasExecutionEngine
        """  # noqa: E501 # FIXME CoP

        # TODO: check that the name is a valid python identifier (and maybe that it is snake_case?)
        ds_type_name = _get_field_details(ds_type, "type").default_value
        if not ds_type_name:
            raise TypeRegistrationError(  # noqa: TRY003 # FIXME CoP
                f"`{ds_type.__name__}` is missing a `type` attribute with an assigned string value"
            )

        # rollback type registrations if exception occurs
        with (
            cls.type_lookup.transaction() as ds_type_lookup,
            ds_type._type_lookup.transaction() as asset_type_lookup,
        ):
            cls._register_assets(ds_type, asset_type_lookup=asset_type_lookup)

            cls._register_datasource(
                ds_type,
                ds_type_name=ds_type_name,
                datasource_type_lookup=ds_type_lookup,
            )

    @classmethod
    def _register_datasource(
        cls,
        ds_type: Type[Datasource],
        ds_type_name: str,
        datasource_type_lookup: TypeLookup,
    ) -> str:
        """
        Register the `Datasource` class and add a factory method for the class on `sources`.
        The method name is pulled from the `Datasource.type` attribute.
        """
        if ds_type in datasource_type_lookup:
            raise TypeRegistrationError(  # noqa: TRY003 # FIXME CoP
                f"'{ds_type_name}' is already a registered typed and there can only be 1 type "
                "for a given name."
            )
        datasource_type_lookup[ds_type] = ds_type_name
        logger.debug(f"'{ds_type_name}' added to `type_lookup`")

        cls._register_add_datasource(ds_type, ds_type_name)
        cls._register_update_datasource(ds_type, ds_type_name)
        cls._register_add_or_update_datasource(ds_type, ds_type_name)
        cls._register_delete_datasource(ds_type, ds_type_name)

        return ds_type_name

    @classmethod
    def _register_add_datasource(cls, ds_type: Type[Datasource], ds_type_name: str):
        method_name = f"add_{ds_type_name}"

        def crud_method_info() -> tuple[CrudMethodType, Type[Datasource]]:
            return CrudMethodType.ADD, ds_type

        cls._register_crud_method(method_name, cls.__doc__, crud_method_info)

    @classmethod
    def _register_update_datasource(cls, ds_type: Type[Datasource], ds_type_name: str):
        method_name = f"update_{ds_type_name}"

        def crud_method_info() -> tuple[CrudMethodType, Type[Datasource]]:
            return CrudMethodType.UPDATE, ds_type

        cls._register_crud_method(
            method_name, f"Updates a {ds_type_name} data source.", crud_method_info
        )

    @classmethod
    def _register_add_or_update_datasource(cls, ds_type: Type[Datasource], ds_type_name: str):
        method_name = f"add_or_update_{ds_type_name}"

        def crud_method_info() -> tuple[CrudMethodType, Type[Datasource]]:
            return CrudMethodType.ADD_OR_UPDATE, ds_type

        cls._register_crud_method(
            method_name,
            f"Adds or updates a {ds_type_name} data source.",
            crud_method_info,
        )

    @classmethod
    def _register_delete_datasource(cls, ds_type: Type[Datasource], ds_type_name: str):
        method_name = f"delete_{ds_type_name}"

        def crud_method_info() -> tuple[CrudMethodType, Type[Datasource]]:
            return CrudMethodType.DELETE, ds_type

        cls._register_crud_method(
            method_name, f"Deletes a {ds_type_name} data source.", crud_method_info
        )

    @classmethod
    def _register_crud_method(
        cls,
        crud_fn_name: str,
        crud_fn_doc: str | None,
        crud_method_info: CrudMethodInfoFn,
    ):
        # We set the name and doc and the crud_method_info because that will be used as a proxy
        # for the real crud method so we can call public_api to generate docs for these dynamically
        # generated methods.
        crud_method_info.__name__ = crud_fn_name
        crud_method_info.__doc__ = crud_fn_doc
        if crud_fn_name in cls.__crud_registry:
            raise TypeRegistrationError(  # noqa: TRY003 # FIXME CoP
                f"'`sources.{crud_fn_name}()` already exists",
            )
        logger.debug(f"Registering data_context.source.{crud_fn_name}()")
        public_api(crud_method_info)
        cls.__crud_registry[crud_fn_name] = crud_method_info

    @classmethod
    def _register_assets(cls, ds_type: Type[Datasource], asset_type_lookup: TypeLookup):
        asset_types: Sequence[Type[DataAsset]] = ds_type.asset_types

        if not asset_types:
            logger.warning(
                f"No `{ds_type.__name__}.asset_types` have be declared for the `Datasource`"
            )

        for t in asset_types:
            if t.__name__.startswith("_"):
                logger.debug(
                    f"{t} is private, assuming not intended as a public concrete type. Skipping registration"  # noqa: E501 # FIXME CoP
                )
                continue
            try:
                asset_type_name = _get_field_details(t, "type").default_value
                if asset_type_name is None:
                    raise TypeError(  # noqa: TRY003, TRY301 # FIXME CoP
                        f"{t.__name__} `type` field must be assigned and cannot be `None`"
                    )
                logger.debug(
                    f"Registering `{ds_type.__name__}` `DataAsset` `{t.__name__}` as '{asset_type_name}'"  # noqa: E501 # FIXME CoP
                )
                asset_type_lookup[t] = asset_type_name
            except (AttributeError, KeyError, TypeError) as bad_field_exc:
                raise TypeRegistrationError(  # noqa: TRY003 # FIXME CoP
                    f"No `type` field found for `{ds_type.__name__}.asset_types` -> `{t.__name__}` unable to register asset type",  # noqa: E501 # FIXME CoP
                ) from bad_field_exc

            cls._bind_asset_factory_method_if_not_present(ds_type, t, asset_type_name)

    @classmethod
    def _bind_asset_factory_method_if_not_present(
        cls,
        ds_type: Type[Datasource],
        asset_type: Type[DataAsset],
        asset_type_name: str,
    ):
        add_asset_factory_method_name = f"add_{asset_type_name}_asset"
        asset_factory_defined: bool = hasattr(ds_type, add_asset_factory_method_name)

        if not asset_factory_defined:
            logger.debug(
                f"No `{add_asset_factory_method_name}()` method found for `{ds_type.__name__}` generating the method..."  # noqa: E501 # FIXME CoP
            )

            def _add_asset_factory(self: Datasource, name: str, **kwargs) -> pydantic.BaseModel:
                # if the Datasource uses a data_connector we need to identify the
                # asset level attributes needed by the data_connector
                # push them to `connect_options` field
                if self.data_connector_type:
                    logger.info(
                        f"'{self.name}' {type(self).__name__} uses {self.data_connector_type.__name__}"  # noqa: E501 # FIXME CoP
                    )
                    connect_options = {
                        k: v
                        for (k, v) in kwargs.items()
                        if k in self.data_connector_type.asset_level_option_keys
                    }
                    if connect_options:
                        logger.info(
                            f"{self.data_connector_type.__name__} connect_options provided -> {list(connect_options.keys())}"  # noqa: E501 # FIXME CoP
                        )
                        for k in connect_options:  # TODO: avoid this extra loop
                            kwargs.pop(k)
                        kwargs["connect_options"] = connect_options
                        # asset_options_type should raise an error if the options are invalid
                        self.data_connector_type.asset_options_type(**connect_options)
                else:
                    connect_options = {}

                asset = asset_type(name=name, **kwargs)
                return self._add_asset(asset, connect_options=connect_options)

            # attr-defined issue
            # https://github.com/python/mypy/issues/12472
            _add_asset_factory.__signature__ = _merge_signatures(  # type: ignore[attr-defined] # FIXME CoP
                _add_asset_factory, asset_type, exclude={"type"}
            )
            _add_asset_factory.__name__ = add_asset_factory_method_name
            setattr(ds_type, add_asset_factory_method_name, _add_asset_factory)

            # NOTE: Please review what this looks like in our Public API docs preview before merging
            _add_asset_factory.__doc__ = DataSourceManager._build_add_asset_docstring(
                asset_type_name
            )

            # add the public api decorator
            public_api(getattr(ds_type, add_asset_factory_method_name))

            if getattr(ds_type, "ADD_READER_METHODS", False):

                def _read_asset_factory(
                    self: Datasource, asset_name: str | None = None, **kwargs
                ) -> Validator:
                    name = asset_name or DEFAULT_PANDAS_DATA_ASSET_NAME
                    asset = asset_type(name=name, **kwargs)
                    self._add_asset(asset)
                    batch_request = asset.build_batch_request()
                    # TODO: raise error if `_data_context` not set
                    return self._data_context.get_validator(batch_request=batch_request)  # type: ignore[union-attr] # self._data_context must be set

                _read_asset_factory.__signature__ = _merge_signatures(  # type: ignore[attr-defined] # FIXME CoP
                    _read_asset_factory, asset_type, exclude={"type"}
                )
                read_asset_factory_method_name = f"read_{asset_type_name}"
                setattr(ds_type, read_asset_factory_method_name, _read_asset_factory)

        else:
            logger.debug(
                f"`{add_asset_factory_method_name}()` already defined `{ds_type.__name__}`"
            )

    @staticmethod
    def _build_add_asset_docstring(asset_type_name: str) -> str:
        article = "an" if asset_type_name[0].lower() in "aeiou" else "a"
        return f"""Add {article} {asset_type_name} asset to the datasource."""

    @property
    def pandas_default(self) -> PandasDatasource:
        from great_expectations.datasource.fluent import PandasDatasource

        datasources = self.all()

        # if a legacy datasource with this name already exists, we try a different name
        existing_datasource = datasources.get(DEFAULT_PANDAS_DATASOURCE_NAME)

        if not existing_datasource:
            return self._data_context.data_sources.add_pandas(name=DEFAULT_PANDAS_DATASOURCE_NAME)

        if isinstance(existing_datasource, PandasDatasource):
            return existing_datasource

        raise DefaultPandasDatasourceError(  # noqa: TRY003 # FIXME CoP
            "Another non-pandas datasource already exists "
            f'with the name: "{DEFAULT_PANDAS_DATASOURCE_NAME}". '
            "Please rename this datasources if you wish "
            "to use the pandas_default `PandasDatasource`."
        )

    @property
    def factories(self) -> List[str]:
        return list(self.__crud_registry.keys())

    def _validate_current_datasource_type(
        self, name: str, datasource_type: Type[Datasource], raise_if_none: bool = True
    ) -> None:
        try:
            current_datasource = self._data_context.data_sources.get(name)
        except KeyError as e:
            if raise_if_none:
                raise ValueError(f"There is no datasource {name} in the data context.") from e  # noqa: TRY003 # FIXME CoP
            current_datasource = None
        if current_datasource and not isinstance(current_datasource, datasource_type):
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"Trying to update datasource {name} but it is not the correct type. "
                f"Expected {datasource_type.__name__} but got {type(current_datasource).__name__}"
            )

    def _datasource_passed_in_as_only_argument(
        self,
        datasource_type: Type[Datasource],
        name_or_datasource: Optional[Union[str, Datasource]],
        **kwargs,
    ) -> Optional[Datasource]:
        """Returns a datasource if one is passed in, otherwise None."""
        from great_expectations.datasource.fluent.interfaces import Datasource

        datasource: Optional[Datasource] = None
        if name_or_datasource and isinstance(name_or_datasource, Datasource):
            if len(kwargs) != 0:
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    f"The datasource must be the sole argument. We also received: {kwargs}"
                )
            datasource = name_or_datasource
        elif name_or_datasource is None and "datasource" in kwargs:
            if len(kwargs) != 1:
                raise ValueError(f"The datasource must be the sole argument. We received: {kwargs}")  # noqa: TRY003 # FIXME CoP
            datasource = kwargs["datasource"]
        if datasource and not isinstance(datasource, datasource_type):
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                f"Trying to modify datasource {datasource.name} but it is not the correct type. "
                f"Expected {datasource_type} but got {type(datasource)}"
            )
        return datasource

    def _datasource_passed_in(
        self,
        datasource_type: Type[Datasource],
        name_or_datasource: Optional[Union[str, Datasource]],
        **kwargs,
    ) -> Optional[Datasource]:
        """Validates the input is a datasource or a set of constructor parameters

        The first argument can be a non-keyword argument. If present it is a datasource
        or the name. If it is None, we expect to find either datasource or name as a kwarg.

        Args:
             datasource_type: The expected type of datasource
             name_or_datasource: Either the datasource or the name of the datasource.

        Returns:
            The passed in datasource or None.

        Raises:
            ValueError: This is raised if a datasource is passed in with additional arguments or
                        a datasource is not passed in and no name argument is present.
        """
        new_datasource = self._datasource_passed_in_as_only_argument(
            datasource_type, name_or_datasource, **kwargs
        )
        if new_datasource:
            return new_datasource
        if (
            name_or_datasource and isinstance(name_or_datasource, str) and "name" not in "kwargs"  # noqa: PLR0133 # FIXME CoP
        ) or (name_or_datasource is None and "name" in kwargs and isinstance(kwargs["name"], str)):
            return None
        raise ValueError(  # noqa: TRY003 # FIXME CoP
            "A datasource object or a name string must be present. The datasource or "
            "name can be passed in as the first and only positional argument or can be"
            "can be passed in as keyword arguments. The arguments we received were: "
            f"positional argument: {name_or_datasource}, kwargs: {kwargs}"
        )

    def create_add_crud_method(
        self,
        datasource_type: Type[Datasource],
        doc_string: str = "",
    ) -> SourceFactoryFn:
        def add_datasource(
            name_or_datasource: Optional[Union[str, Datasource]] = None, **kwargs
        ) -> Datasource:
            # Because of the precedence of `or` and `if`, these grouping paranthesis are necessary.
            datasource = (
                self._datasource_passed_in(datasource_type, name_or_datasource, **kwargs)
            ) or (
                datasource_type(name=name_or_datasource, **kwargs)
                if name_or_datasource
                else datasource_type(**kwargs)
            )
            logger.debug(f"Adding {datasource_type} with {datasource.name}")
            datasource._data_context = self._data_context
            datasource.test_connection()
            datasource = self._data_context._add_fluent_datasource(datasource)

            return datasource

        add_datasource.__doc__ = doc_string
        # attr-defined issue https://github.com/python/mypy/issues/12472
        add_datasource.__signature__ = _merge_signatures(  # type: ignore[attr-defined] # FIXME CoP
            add_datasource,
            datasource_type,
            exclude={"type", "assets"},
            return_type=datasource_type,
        )
        return add_datasource

    def create_update_crud_method(
        self,
        datasource_type: Type[Datasource],
        doc_string: str = "",
    ) -> SourceFactoryFn:
        def update_datasource(
            name_or_datasource: Optional[Union[str, Datasource]] = None, **kwargs
        ) -> Datasource:
            # circular import
            from great_expectations.datasource.fluent.interfaces import Datasource

            updated_datasource = (
                self._datasource_passed_in(datasource_type, name_or_datasource, **kwargs)
            ) or (
                datasource_type(name=name_or_datasource, **kwargs)
                if name_or_datasource
                else datasource_type(**kwargs)
            )

            datasource_name: str = updated_datasource.name
            logger.debug(f"Updating {datasource_type} with {datasource_name}")
            self._validate_current_datasource_type(
                datasource_name,
                datasource_type,
            )

            # preserve any pre-existing id for usage with cloud
            id_: uuid.UUID | None = getattr(self.all().get(datasource_name), "id", None)
            if id_:
                updated_datasource.id = id_

            updated_datasource._data_context = self._data_context
            updated_datasource.test_connection()
            return_obj = self._data_context._update_fluent_datasource(datasource=updated_datasource)
            assert isinstance(return_obj, Datasource)
            return return_obj

        update_datasource.__doc__ = doc_string
        # attr-defined issue https://github.com/python/mypy/issues/12472
        update_datasource.__signature__ = _merge_signatures(  # type: ignore[attr-defined] # FIXME CoP
            update_datasource,
            datasource_type,
            exclude={"type", "assets"},
            return_type=datasource_type,
        )
        return update_datasource

    def create_add_or_update_crud_method(
        self,
        datasource_type: Type[Datasource],
        doc_string: str = "",
    ) -> SourceFactoryFn:
        def add_or_update_datasource(
            name_or_datasource: Optional[Union[str, Datasource]] = None, **kwargs
        ) -> Datasource:
            # circular import
            from great_expectations.datasource.fluent.interfaces import Datasource

            new_datasource = (
                self._datasource_passed_in(datasource_type, name_or_datasource, **kwargs)
            ) or (
                datasource_type(name=name_or_datasource, **kwargs)
                if name_or_datasource
                else datasource_type(**kwargs)
            )

            # if new_datasource is None that means name is defined as name_or_datasource or as a kwarg  # noqa: E501 # FIXME CoP
            datasource_name: str = new_datasource.name
            logger.debug(f"Adding or updating {datasource_type.__name__} with '{datasource_name}'")
            self._validate_current_datasource_type(
                datasource_name, datasource_type, raise_if_none=False
            )

            # preserve any pre-existing id for usage with cloud
            id_: uuid.UUID | None = getattr(self.all().get(datasource_name), "id", None)
            if id_:
                new_datasource.id = id_

            new_datasource._data_context = self._data_context
            new_datasource.test_connection()
            if datasource_name in self.all():
                return_obj = self._data_context._update_fluent_datasource(datasource=new_datasource)
            else:
                return_obj = self._data_context._add_fluent_datasource(datasource=new_datasource)
            assert isinstance(return_obj, Datasource)
            return return_obj

        add_or_update_datasource.__doc__ = doc_string
        # attr-defined issue https://github.com/python/mypy/issues/12472
        add_or_update_datasource.__signature__ = _merge_signatures(  # type: ignore[attr-defined] # FIXME CoP
            add_or_update_datasource,
            datasource_type,
            exclude={"type", "assets"},
            return_type=datasource_type,
        )
        return add_or_update_datasource

    def create_delete_crud_method(
        self,
        datasource_type: Type[Datasource],
        doc_string: str = "",
    ) -> Callable[[str], None]:
        def delete_datasource(name: str) -> None:
            logger.debug(f"Delete {datasource_type} with {name}")
            self._validate_current_datasource_type(name, datasource_type)
            self._data_context._delete_fluent_datasource(name=name)
            self._data_context._save_project_config()

        delete_datasource.__doc__ = doc_string
        # attr-defined issue https://github.com/python/mypy/issues/12472
        delete_datasource.__signature__ = inspect.signature(delete_datasource)  # type: ignore[attr-defined] # FIXME CoP
        return delete_datasource

    @public_api
    def delete(self, name: str) -> None:
        """
        Deletes a datasource by name.

        Args:
            name: The name of the given datasource.
        """
        self._data_context.delete_datasource(name=name)

    @public_api
    def all(self) -> DatasourceDict:
        """Get all Datasources."""
        return self._data_context._datasources

    @public_api
    def get(self, name: str) -> Datasource:
        """Get a Datasource from the collection by name.

        Parameters:
            name: Name of Datasource to get

        Raises:
            KeyError: when Datasource is not found.
        """
        return self.all()[name]

    def __getattr__(self, attr_name: str):
        try:
            crud_method_info = self.__crud_registry[attr_name]
            crud_method_type, datasource_type = crud_method_info()
            docstring = crud_method_info.__doc__ or ""
            if crud_method_type == CrudMethodType.ADD:
                return self.create_add_crud_method(datasource_type, docstring)
            elif crud_method_type == CrudMethodType.UPDATE:
                return self.create_update_crud_method(datasource_type, docstring)
            elif crud_method_type == CrudMethodType.ADD_OR_UPDATE:
                return self.create_add_or_update_crud_method(datasource_type, docstring)
            elif crud_method_type == CrudMethodType.DELETE:
                # deprecated-v0.17.2
                warnings.warn(
                    f"`{attr_name}` is deprecated as of v0.17.2 and will be removed in v0.19. Please use `.sources.delete` moving forward.",  # noqa: E501 # FIXME CoP
                    DeprecationWarning,
                )
                return self.create_delete_crud_method(datasource_type, docstring)
            else:
                raise TypeRegistrationError(  # noqa: TRY003 # FIXME CoP
                    f"Unknown crud method registered for {attr_name} with type {crud_method_type}"
                )
        except KeyError as e:
            raise AttributeError(f"No crud method '{attr_name}' in {self.factories}") from e  # noqa: TRY003 # FIXME CoP

    @override
    def __dir__(self) -> List[str]:
        """Preserves autocompletion for dynamic attributes."""
        return [*self.factories, *super().__dir__()]


def _iter_all_registered_types(
    include_datasource: bool = True, include_data_asset: bool = True
) -> Generator[tuple[str, Type[Datasource] | Type[DataAsset]], None, None]:
    """
    Iterate through all registered Datasource and DataAsset types.
    Returns tuples of the registered type name and the actual type/class.
    """
    for ds_name in DataSourceManager.type_lookup.type_names():
        ds_type: Type[Datasource] = DataSourceManager.type_lookup[ds_name]
        if include_datasource:
            yield ds_name, ds_type

        if include_data_asset:
            for asset_name in ds_type._type_lookup.type_names():
                asset_type: Type[DataAsset] = ds_type._type_lookup[asset_name]
                yield asset_name, asset_type



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Any, Dict, List, Optional, TypeVar, Union

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core import ExpectationSuite
from great_expectations.core.expectation_suite import ExpectationSuiteSchema
from great_expectations.data_context.cloud_constants import GXCloudRESTResource
from great_expectations.data_context.store.database_store_backend import (
    DatabaseStoreBackend,
)
from great_expectations.data_context.store.store import Store
from great_expectations.data_context.store.tuple_store_backend import TupleStoreBackend
from great_expectations.data_context.types.refs import GXCloudResourceRef
from great_expectations.data_context.types.resource_identifiers import (
    ExpectationSuiteIdentifier,
    GXCloudIdentifier,
)
from great_expectations.util import (
    filter_properties_dict,
)

if TYPE_CHECKING:
    from great_expectations.data_context.data_context.abstract_data_context import (
        AbstractDataContext,
    )
    from great_expectations.expectations.expectation import Expectation

    _TExpectation = TypeVar("_TExpectation", bound=Expectation)


class ExpectationConfigurationDTO(pydantic.BaseModel):
    class Config:
        extra = pydantic.Extra.ignore

    id: str
    type: str
    rendered_content: List[dict] = pydantic.Field(default_factory=list)
    kwargs: dict
    meta: Union[dict, None]
    description: Union[str, None]
    expectation_context: Union[dict, None]


class ExpectationSuiteDTO(pydantic.BaseModel):
    """Capture known fields from a serialized ExpectationSuite."""

    class Config:
        extra = pydantic.Extra.ignore

    name: str
    id: str
    expectations: List[ExpectationConfigurationDTO]
    meta: Union[dict, None]
    notes: Union[str, None]


class ExpectationsStore(Store):
    """
    An Expectations Store provides a way to store Expectation Suites accessible to a Data Context.
    """

    _key_class = ExpectationSuiteIdentifier

    def __init__(
        self,
        store_backend: dict | None = None,
        runtime_environment: dict | None = None,
        store_name: str = "no_store_name",
        data_context: AbstractDataContext | None = None,
    ) -> None:
        self._expectationSuiteSchema = ExpectationSuiteSchema()
        self._data_context = data_context

        store_backend_class = self._determine_store_backend_class(store_backend)
        # Store Backend Class was loaded successfully; verify that it is of a correct subclass.
        if store_backend:
            if issubclass(store_backend_class, TupleStoreBackend):
                # Provide defaults for this common case
                store_backend["filepath_suffix"] = store_backend.get("filepath_suffix", ".json")
            elif issubclass(store_backend_class, DatabaseStoreBackend):
                # Provide defaults for this common case
                store_backend["table_name"] = store_backend.get(
                    "table_name", "ge_expectations_store"
                )
                store_backend["key_columns"] = store_backend.get(
                    "key_columns", ["expectation_suite_name"]
                )

        super().__init__(
            store_backend=store_backend,
            runtime_environment=runtime_environment,
            store_name=store_name,
        )

        # Gather the call arguments of the present function (include the "module_name" and add the "class_name"), filter  # noqa: E501 # FIXME CoP
        # out the Falsy values, and set the instance "_config" variable equal to the resulting dictionary.  # noqa: E501 # FIXME CoP
        self._config = {
            "store_backend": store_backend,
            "runtime_environment": runtime_environment,
            "store_name": store_name,
            "module_name": self.__class__.__module__,
            "class_name": self.__class__.__name__,
        }
        filter_properties_dict(properties=self._config, clean_falsy=True, inplace=True)

    @override
    @classmethod
    def gx_cloud_response_json_to_object_dict(cls, response_json: dict) -> dict:
        """
        This method takes full json response from GX cloud and outputs a dict appropriate for
        deserialization into a GX object
        """
        suite_data: Dict
        # if only the expectation_suite_name is passed, a list will be returned
        if isinstance(response_json["data"], list):
            if len(response_json["data"]) == 1:
                suite_data = response_json["data"][0]
            else:
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    "More than one Expectation Suite was found with the expectation_suite_name."
                )
        else:
            suite_data = response_json["data"]

        return cls._convert_raw_json_to_object_dict(suite_data)

    @override
    @staticmethod
    def _convert_raw_json_to_object_dict(data: dict[str, Any]) -> dict[str, Any]:
        # Cloud backend adds a default result format type of None, so ensure we remove it:
        for expectation in data.get("expectations", []):
            kwargs = expectation["kwargs"]
            if "result_format" in kwargs and kwargs["result_format"] is None:
                kwargs.pop("result_format")

        suite_dto = ExpectationSuiteDTO.parse_obj(data)
        return suite_dto.dict()

    def add_expectation(self, suite: ExpectationSuite, expectation: _TExpectation) -> _TExpectation:
        suite_identifier, fetched_suite = self._refresh_suite(suite)

        # we need to find which ID has been added by the backend
        old_ids = {exp.id for exp in fetched_suite.expectations}

        if self.cloud_mode:
            expectation.id = None  # flag this expectation as new for the backend
        else:
            expectation.id = str(uuid.uuid4())
        fetched_suite.expectations.append(expectation)

        self.update(key=suite_identifier, value=fetched_suite)
        if self.cloud_mode:
            # since update doesn't return the object we need (here), we refetch the suite
            suite_identifier, fetched_suite = self._refresh_suite(suite)
            new_ids = [exp.id for exp in fetched_suite.expectations if exp.id not in old_ids]
            if len(new_ids) > 1:
                # edge case: suite has been changed remotely, and one or more new expectations
                #            have been added. Since the store doesn't return the updated object,
                #            we have no reliable way to know which new ID belongs to this expectation,  # noqa: E501 # FIXME CoP
                #            so we raise an exception and ask the user to refresh their suite.
                #            The Expectation should have been successfully added to the suite.
                raise RuntimeError(  # noqa: TRY003 # FIXME CoP
                    "Expectation was added, however this ExpectationSuite is out of sync with the Cloud backend. "  # noqa: E501 # FIXME CoP
                    f'Please fetch the latest state of this suite by calling `context.suites.get(name="{suite.name}")`.'  # noqa: E501 # FIXME CoP
                )
            elif len(new_ids) == 0:
                # edge case: this is an unexpected state - if the cloud backend failed to add the expectation,  # noqa: E501 # FIXME CoP
                #            it should have already raised an exception.
                raise RuntimeError("Unknown error occurred and Expectation was not added.")  # noqa: TRY003 # FIXME CoP
            else:
                new_id = new_ids[0]
            expectation.id = new_id
        return expectation

    def update_expectation(self, suite: ExpectationSuite, expectation: Expectation) -> Expectation:
        suite_identifier, fetched_suite = self._refresh_suite(suite)

        if expectation.id not in {exp.id for exp in fetched_suite.expectations}:
            raise KeyError("Cannot update Expectation because it was not found.")  # noqa: TRY003 # FIXME CoP

        for i, old_expectation in enumerate(fetched_suite.expectations):
            if old_expectation.id == expectation.id:
                fetched_suite.expectations[i] = expectation
                break

        self.update(key=suite_identifier, value=fetched_suite)
        # we don't expect the backend to have made changes to the Expectation,
        # so we don't update its in-memory reference.

        return expectation

    def delete_expectation(self, suite: ExpectationSuite, expectation: Expectation) -> Expectation:
        suite_identifier, suite = self._refresh_suite(suite)

        if expectation.id not in {exp.id for exp in suite.expectations}:
            raise KeyError("Cannot delete Expectation because it was not found.")  # noqa: TRY003 # FIXME CoP

        for i, old_expectation in enumerate(suite.expectations):
            if old_expectation.id == expectation.id:
                del suite.expectations[i]
                break

        self.update(key=suite_identifier, value=suite)
        return expectation

    def _refresh_suite(
        self, suite
    ) -> tuple[Union[GXCloudIdentifier, ExpectationSuiteIdentifier], ExpectationSuite]:
        """Get the latest state of an ExpectationSuite from the backend."""
        suite_identifier = self.get_key(name=suite.name, id=suite.id)
        suite_dict = self.get(key=suite_identifier)
        suite = ExpectationSuite(**suite_dict)
        return suite_identifier, suite

    def _add(self, key, value, **kwargs):  # type: ignore[explicit-override] # FIXME
        if not self.cloud_mode:
            # this logic should move to the store backend, but is implemented here for now
            value: ExpectationSuite = self._add_ids_on_create(value)
        try:
            result = super()._add(key=key, value=value, **kwargs)
            if self.cloud_mode:
                # cloud backend has added IDs, so we update our local state to be in sync
                assert isinstance(result, GXCloudResourceRef)
                suite_kwargs = self.deserialize(
                    self.gx_cloud_response_json_to_object_dict(result.response)
                )
                cloud_suite = ExpectationSuite(**suite_kwargs)
                value = self._add_cloud_ids_to_local_suite_and_expectations(
                    local_suite=value,
                    cloud_suite=cloud_suite,
                )
            return result
        except gx_exceptions.StoreBackendError as exc:
            raise gx_exceptions.ExpectationSuiteError(  # noqa: TRY003 # FIXME CoP
                f"An error occurred while trying to save ExpectationSuite: {exc.message}"
            ) from exc

    def _update(self, key, value, **kwargs):  # type: ignore[explicit-override] # FIXME
        if not self.cloud_mode:
            # this logic should move to the store backend, but is implemented here for now
            value: ExpectationSuite = self._add_ids_on_update(value)
        try:
            result = super()._update(key=key, value=value, **kwargs)

            if self.cloud_mode:
                # cloud backend has added IDs, so we update our local state to be in sync
                assert isinstance(result, GXCloudResourceRef)
                suite_kwargs = self.deserialize(
                    self.gx_cloud_response_json_to_object_dict(result.response)
                )
                cloud_suite = ExpectationSuite(**suite_kwargs)
                value = self._add_cloud_ids_to_local_suite_and_expectations(
                    local_suite=value,
                    cloud_suite=cloud_suite,
                )
        except gx_exceptions.StoreBackendError as e:
            # todo: this generic error clobbers more informative errors coming from the store

            raise gx_exceptions.ExpectationSuiteNotAddedError(name=value.name) from e

    def _add_ids_on_create(self, suite: ExpectationSuite) -> ExpectationSuite:
        """This method handles adding IDs to suites and expectations for non-cloud backends.
        In the future, this logic should be the responsibility of each non-cloud backend.
        """
        suite["id"] = str(uuid.uuid4())
        if isinstance(suite, ExpectationSuite):
            for expectation in suite.expectations:
                expectation.id = str(uuid.uuid4())
        else:
            for expectation in suite["expectations"]:
                expectation["id"] = str(uuid.uuid4())

        return suite

    def _add_ids_on_update(self, suite: ExpectationSuite) -> ExpectationSuite:
        """This method handles adding IDs to suites and expectations for non-cloud backends.
        In the future, this logic should be the responsibility of each non-cloud backend.
        """

        if not suite.id:
            suite.id = str(uuid.uuid4())

        # enforce that every ID in this suite is unique
        expectation_ids = [exp.id for exp in suite.expectations if exp.id]
        if len(expectation_ids) != len(set(expectation_ids)):
            raise RuntimeError("Expectation IDs must be unique within a suite.")  # noqa: TRY003 # FIXME CoP

        for expectation in suite.expectations:
            if not expectation.id:
                expectation.id = str(uuid.uuid4())
        return suite

    def _add_cloud_ids_to_local_suite_and_expectations(
        self, local_suite: ExpectationSuite, cloud_suite: ExpectationSuite
    ) -> ExpectationSuite:
        if not local_suite.id:
            local_suite.id = cloud_suite.id
        # We replace local expectations with those returned from the backend
        # so remote changes are reflected in the in-memory ExpectationSuite.
        # Note that the parent Suite of these Expectations is actually `cloud_suite`,
        # since we aren't using the public ExpectationSuite API to add the Expectations.
        # This means that `Expectation._save_callback` is provided by a different copy of the
        # same ExpectationSuite.
        local_suite.expectations = [expectation for expectation in cloud_suite.expectations]
        return local_suite

    @override
    def get(self, key) -> dict:
        return super().get(key)  # type: ignore[return-value] # FIXME CoP

    @override
    def _validate_key(  # type: ignore[override] # FIXME CoP
        self, key: ExpectationSuiteIdentifier | GXCloudIdentifier
    ) -> None:
        if isinstance(key, GXCloudIdentifier) and not key.id and not key.resource_name:
            raise ValueError(  # noqa: TRY003 # FIXME CoP
                "GXCloudIdentifier for ExpectationsStore must contain either "
                "an id or a resource_name, but neither are present."
            )
        return super()._validate_key(key=key)

    def serialize(self, value):  # type: ignore[explicit-override] # FIXME
        if self.cloud_mode:
            # GXCloudStoreBackend expects a json str
            val = self._expectationSuiteSchema.dump(value)
            return val
        return self._expectationSuiteSchema.dumps(value, indent=2, sort_keys=True)

    def deserialize(self, value):  # type: ignore[explicit-override] # FIXME
        if isinstance(value, dict):
            return self._expectationSuiteSchema.load(value)
        elif isinstance(value, str):
            return self._expectationSuiteSchema.loads(value)
        else:
            raise TypeError(f"Cannot deserialize value of unknown type: {type(value)}")  # noqa: TRY003 # FIXME CoP

    def deserialize_suite_dict(self, suite_dict: dict) -> ExpectationSuite:
        suite = ExpectationSuite(**suite_dict)
        if suite._include_rendered_content:
            suite.render()
        return suite

    def get_key(
        self, name: str, id: Optional[str] = None
    ) -> GXCloudIdentifier | ExpectationSuiteIdentifier:
        """Given a name and optional ID, build the correct key for use in the ExpectationsStore."""
        key: GXCloudIdentifier | ExpectationSuiteIdentifier
        if self.cloud_mode:
            key = GXCloudIdentifier(
                resource_type=GXCloudRESTResource.EXPECTATION_SUITE,
                id=id,
                resource_name=name,
            )
        else:
            key = ExpectationSuiteIdentifier(name=name)
        return key
